import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot

@validate_arguments(config=dict(arbitrary_types_allowed=True))
def khtrack(
    df:pd.DataFrame, 
    kh:Union[str,list]=None, 
    ylims:Tuple[float,float]=None,
    kh_lims:Tuple[float,float]=(0,1),
    dtick:bool=False,
    fill:bool=True, 
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    fontsize: int=8,
    grid_numbers : list = [11,51],
    steps: list  = None,
    kh_kw={},
    corr_kw={},
    fill_kh_kw={},
    kh_colormap:str='gray',
    depth_ref:str='md'
):
    list_axes = []
    hax=ax or plt.gca()
    
    def_kh_kw = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }
    
    for (k,v) in def_kh_kw.items():
        if k not in kh_kw:
            kh_kw[k]=v
            
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_corr_kw.items():
        if k not in corr_kw:
            corr_kw[k]=v

    def_fill_kh_kw = {
    'color': (0.5,0.5,0.5),
    }    
    for (k,v) in def_fill_kh_kw.items():
        if k not in fill_kh_kw:
            fill_kh_kw[k]=v

    depth = df.index if depth_ref=='md' else df[depth_ref]
    if kh is not None:
        if isinstance(kh,str):
            hax.plot(df[kh],depth,**kh_kw)   #Plotting
        elif isinstance(kh,list):
            cmap = mpl.cm.get_cmap(kh_colormap,len(kh))
            for i,g in enumerate(kh):
                kh_kw['color']=cmap(i)
                hax.plot(df[g],depth,**kh_kw)
    
    if ylims==None: #Depth Limits
        ylims=[depth.min(),depth.max()]

    hax.set_ylim(list(ylims))
    hax.set_xlim(list(kh_lims))

    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
    
    hax.set_xlim(list(kh_lims))
    hax.set_xlabel("Kh Norm")
    hax.set_xticks(np.linspace(0,1,4))
    hax.set_xticklabels(np.round(np.linspace(0,1,4),decimals=2))
    hax.xaxis.tick_top()
    hax.xaxis.set_label_position("top")
    hax.tick_params("both",labelsize=fontsize)
    hax.set_yticks(mayor_grid)
    hax.set_yticks(minor_grid,minor=True)        
    if dtick==True:
        hax.set_yticklabels(mayor_grid)
    else:
        hax.set_yticklabels([])

    if fill==True:
        hax.fill_betweenx(depth,kh_lims[0],df[kh],**fill_kh_kw)
        
    #Add formations tops
    if formations is not None:
        hax = formations_plot(
            formations,
            depth_ref,
            ylims,
            kh_lims,
            hax,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        hax = formations_plot(
            units,
            depth_ref,
            ylims,
            kh_lims,
            hax,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        hax = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            kh_lims,
            hax,
            config=perforations_kw
        )
        
    if correlations is not None:
        hax = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            kh_lims,
            hax,
            config=correlations_kw
        )
        
    list_axes.append(hax)
    return hax


import numpy as np 
import pandas as pd 
from scipy.interpolate import interp1d
from scipy.stats import hmean, gmean
from pydantic import BaseModel, Field, validator
from enum import Enum

class PetrophysicsEstimator(BaseModel):
    description: str = Field(None)

def vshale_gr(gr_curve,gr_sand,gr_shale,type='linear'):
    """vshale_gr [summary]

    Parameters
    ----------
    gr_curve : [type]
        [description]
    gr_sand : [type]
        [description]
    gr_shale : [type]
        [description]
    type : str, optional
        [description], by default 'linear'

    Returns
    -------
    [type]
        [description]

    Raises
    ------
    ValueError
        [description]
    """
    gr_curve=np.atleast_1d(gr_curve)
    gr_sand=np.atleast_1d(gr_sand)
    gr_shale=np.atleast_1d(gr_shale)
    igr=(gr_curve-gr_sand)/(gr_shale-gr_sand)
    igr[igr < 0.0] = 0.0
    igr[igr > 1.0] = 1.0
    #https://www.geoloil.com/VshModels.php
    if type == 'linear':
        vsh = igr
    elif type == 'clavier':
        vsh = 1.7 - np.sqrt(3.38 - np.power(igr+0.7,2))
    elif type == 'stieber':
        vsh = igr/(3-2*igr)
    elif type == 'larionov_old':
        vsh = 0.33 * (np.power(2,2*igr)-1)
    elif type == 'larionov_tertiary':
        vsh = 0.083 * (np.power(2,3.7*igr)-1)
    else:
        raise ValueError(f'method especified [ {type} ] does not exist')

    return vsh


class VshGrEnum(str,Enum):
    linear = 'linear'
    clavier = 'clavier'
    stieber = 'stieber'
    larionov_old = 'larionov_old'
    larionov_tertiary = 'larionov_tertiary'
    
class VshaleGr(PetrophysicsEstimator):
    gr_sand: float = Field(None, gt=0)
    gr_shale: float = Field(None, gt=0)
    method: VshGrEnum = Field(VshGrEnum.linear)
    col_name: str = 'vsh_gr'
    gr_curve: str = Field(None)
    
    @validator(gr_shale,check_fields=False)
    def check_grshale_gt_grsand(cls,v,values):
        assert v > values['gr_sand'], f"gr_shale {v} must be greater than gr_sand {values['gr_sand']}"
        return v
    
    def forward(self,gr):
        return vshale_gr(gr,self.gr_sand,self.gr_shale,type=self.method.value)
    
    def forward_log(self,df):
        vsh = vshale_gr(df[self.gr_curve],self.gr_sand,self.gr_shale,type=self.method.value)
        return pd.Series(vsh,index=df.index,name=self.col_name)
    
#-----------------------------------------------------------------------------------
def vshale_dn(rho_curve, ntr_curve, rho_ma=2.65, rho_f=1.0, hi_shl=0.46,rho_shl=2.43):
    """vshale_dn [summary]

    Parameters
    ----------
    rho_curve : [type]
        [description]
    ntr_curve : [type]
        [description]
    rho_ma : float, optional
        [description], by default 2.65
    rho_f : float, optional
        [description], by default 1.0
    hi_shl : float, optional
        [description], by default 0.46
    rho_shl : float, optional
        [description], by default 2.43

    Returns
    -------
    [type]
        [description]
    """
    rho_curve= np.atleast_1d(rho_curve)
    ntr_curve= np.atleast_1d(ntr_curve)
    rho_ma = np.atleast_1d(rho_ma)
    rho_f = np.atleast_1d(rho_f)
    hi_shl = np.atleast_1d(hi_shl)
    rho_shl = np.atleast_1d(rho_shl)
    vsh = (rho_curve - rho_ma + ntr_curve*(rho_ma-rho_f))/(rho_shl - rho_ma + hi_shl*(rho_ma-rho_f))
    vsh[vsh < 0.0] = 0.0
    vsh[vsh > 1.0] = 1.0
    return vsh

class VshaleDN(PetrophysicsEstimator):
    rho_matrix: float = Field(2.65, gt=0)
    rho_fluid: float = Field(1.0, gt=0)
    hi_shale: float = Field(0.46)
    rho_shale : float = Field(2.43, gt=0)
    col_name: str = 'vsh_dn'
    rho_curve: str = Field(None)
    ntr_curve: str = Field(None)
       
    def forward(self,rho,ntr):
        return vshale_dn(
            rho, 
            ntr, 
            rho_ma=self.rho_matrix, 
            rho_f=self.rho_fluid, 
            hi_shl=self.hi_shale,
            rho_shl=self.rho_shale
        )
        
    def forward_log(self,df):
        return vshale_dn(
            df[self.rho_curve], 
            df[self.ntr_curve], 
            rho_ma=self.rho_matrix, 
            rho_f=self.rho_fluid, 
            hi_shl=self.hi_shale,
            rho_shl=self.rho_shale
        )
        
    

#-----------------------------------------------------------------------------------

def phi_rho(rho_curve,rho_ma=2.65,rho_f=1.0):
    """phi_rho [summary]

    Parameters
    ----------
    rho_curve : [type]
        [description]
    rho_ma : float, optional
        [description], by default 2.65
    rho_f : float, optional
        [description], by default 1.0

    Returns
    -------
    [type]
        [description]
    """
    rho_curve=np.atleast_1d(rho_curve)
    rho_ma=np.atleast_1d(rho_ma)
    rho_f=np.atleast_1d(rho_f)
    phi_rho_curve=(rho_ma-rho_curve)/(rho_ma-rho_f)
    phi_rho_curve[phi_rho_curve < 0.0] = 0.0
    phi_rho_curve[phi_rho_curve > 1.0] = 1.0
    return phi_rho_curve

class PhiRho(PetrophysicsEstimator):
    rho_matrix: float = Field(2.65, gt=0)
    rho_fluid: float = Field(1.0,gt=0)
    col_name: str = 'phi_rho'
    rho_curve: str = Field(None)
    
    def forward(self,x):
        return phi_rho(x, rho_ma = self.rho_matrix, rho_f=self.rho_fluid)
    
    def forward_log(self,df):
        return phi_rho(df[self.rho_curve], rho_ma = self.rho_matrix, rho_f=self.rho_fluid)

#-----------------------------------------------------------------------------------

def phie(phi_curve,vsh_curve):
    """phie [summary]

    Parameters
    ----------
    phi_curve : [type]
        [description]
    vsh_curve : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    phi_curve=np.atleast_1d(phi_curve)
    vsh_curve=np.atleast_1d(vsh_curve)
    phie_curve=phi_curve*(1 -vsh_curve)
    phie_curve[phie_curve < 0.0] = 0.0
    phie_curve[phie_curve > 0.3] = 0.3
    return phie_curve

class Phie(PetrophysicsEstimator):
    col_name: str = 'phie'
    phi_curve: str = Field(None)
    vsh_curve: str = Field(None)
    
    def forward(self,phi,vsh):
        return phie(phi,vsh)
    
    def forward_log(self,df):
        return phie(
            df[self.phi_curve],
            df[self.vsh_curve]
        )

#-----------------------------------------------------------------------------------

class PhiaMethodEnum(str,Enum):
    arithmetic = 'arithmetic'
    geometric = 'geometric'
    harmonic = 'harmonic'
    

def phia(phi_rho_curve, ntr_curve, method='geometric'):
    """phia [summary]

    Parameters
    ----------
    phi_rho_curve : [type]
        [description]
    ntr_curve : [type]
        [description]
    method : str, optional
        [description], by default 'geometric'

    Returns
    -------
    [type]
        [description]
    """
    phi_rho_curve = np.atleast_1d(phi_rho_curve)
    ntr_curve = np.atleast_1d(ntr_curve)
    c = np.stack([phi_rho_curve,ntr_curve], axis=1)
    if method == 'mean':
        phia_curve = np.mean(c,axis=1)
    elif method== 'geometric':
        phia_curve = gmean(c,axis=1)
    elif method=='harmonic':
        phia_curve = hmean(c,axis=1)
    return phia_curve

class PhiAverage(PetrophysicsEstimator):
    method: PhiaMethodEnum = Field(PhiaMethodEnum.geometric.value)
    col_name: str = 'phia'
    phi1_curve: str = Field(None)
    phi2_curve: str = Field(None)
    
    def forward(self,phi1,phi2):
        return phia(phi1,phi2,method=self.method)
    
    def forward_log(self,df):
        return phia(
            df[self.phi1_curve],
            df[self.phi2_curve],
            method=self.method
        )    
        

#-----------------------------------------------------------------------------------

class SwMethodsEnum(str,Enum):
    archie = 'archie'
    smdx = 'smdx'
    indo = 'indo'
    fertl = 'fertl'

def sw(rt_curve,phi_curve,rw,vsh_curve=None,a=0.62,m=2.15,n=2,rsh=4.0,alpha=0.3,method="archie"):
    """sw [summary]

    Parameters
    ----------
    rt_curve : [type]
        [description]
    phi_curve : [type]
        [description]
    rw : [type]
        [description]
    vsh_curve : [type], optional
        [description], by default None
    a : float, optional
        [description], by default 0.62
    m : float, optional
        [description], by default 2.15
    n : int, optional
        [description], by default 2
    rsh : float, optional
        [description], by default 4.0
    alpha : float, optional
        [description], by default 0.3
    method : str, optional
        [description], by default "archie"

    Returns
    -------
    [type]
        [description]
    """
    a=np.atleast_1d(a)
    m=np.atleast_1d(m)
    n=np.atleast_1d(n)
    vsh = np.atleast_1d(vsh_curve) if vsh_curve is not None else None
    rsh=np.atleast_1d(rsh)
    alpha=np.atleast_1d(alpha)
    rt=np.atleast_1d(rt_curve)
    phi = np.atleast_1d(phi_curve)
    rw=np.atleast_1d(rw)
    if method == "archie":
        sw_curve=np.power(((a*rw)/(rt*np.power(phi,m))),1/n)
    elif method == "smdx": #https://www.spec2000.net/14-sws.htm
        C=((1-vsh)*a*rw)/np.power(phi,m)
        D=C*vsh/(2 * rsh)
        E=C/rt
        sw_curve=np.power(np.sqrt(D**2 + E) - D, 2/n)
    elif method == "indo":
        #https://geoloil.com/Indonesia_SW.php
        #A=np.sqrt(1 /rt)
        #B=(np.power(vsh,(1 -(vsh/2)))/np.sqrt(rsh))
        #C=np.sqrt(np.power(phi,m)/(a*rw))
        #sw_curve=np.power((A/(B+C)),2/n)

        #http://nafta.wiki/display/GLOSSARY/Indonesia+Model+%28Poupon-Leveaux%29+@model
        A_inv = 1 + np.sqrt((np.power(vsh,2-vsh)*rw)/(phi*rsh))
        A = 1/A_inv
        sw_curve = np.power((A*rw)/(np.power(phi,m)*rt),1/n)
    elif method == "fertl":
        A=np.power(phi,-m/2)
        B=(a*rw)/rt
        C=((alpha*vsh)/2)**2
        sw_curve=A*((np.sqrt(B+C))-np.sqrt(C))
    sw_curve[sw_curve < 0.0] = 0.0
    sw_curve[sw_curve > 1.0] = 1.0
    
    return sw_curve

class Sw(PetrophysicsEstimator):
    a: float = Field(0.62)
    m: float = Field(2.15)
    n: float = Field(2.0)
    rsh: float = Field(4.0)
    alpha: float = Field(0.3) 
    method: SwMethodsEnum = Field(SwMethodsEnum.archie)
    col_name: str = 'sw'
    rt_curve: str = Field(None)
    phi_curve: str = Field(None)
    vsh_curve: str = Field(None)
    rw: float = Field(None)
    
    def forward(self,rt,phi,rw=None,vsh=None):
        if rw is None:
            rw = self.rw
        return sw(
            rt,
            phi,
            rw,
            vsh_curve=vsh,
            a = self.a,
            m = self.m,
            n = self.n,
            rsh = self.rsh,
            alpha = self.alpha,
            method = self.method.value
        )

    def forward_log(self,df):
        return sw(
            df[self.vsh_curve],
            df[self.phi_curve],
            self.rw,
            vsh_curve = None if self.vsh_curve is None else df[self.vsh_curve],
            a = self.a,
            m = self.m,
            n = self.n,
            rsh = self.rsh,
            alpha = self.alpha,
            method = self.method.value
        )
        
def depth_temperature(depth, surface_temperature=77 ,gradient=1):
    """depth_temperature [summary]

    Parameters
    ----------
    depth : [type]
        [description]
    surface_temperature : int, optional
        [description], by default 77
    gradient : int, optional
        [description], by default 1

    Returns
    -------
    [type]
        [description]
    """
    depth = np.atleast_1d(depth)
    
    t = (gradient/100) * depth + surface_temperature 
    return t

class DepthTemperature(PetrophysicsEstimator):
    gradient: float = Field(1.0, gt=0)
    surface_temperature: float = Field(60.0, gt=0)
    col_name: str = 'temperature'
    depth_curve: str = Field(None)
    
    def forward(self,depth):
        return depth_temperature(
            depth, 
            self.surface_temperature,
            self.gradient
        )
        
    def forward_log(self,df):
        return depth_temperature(
            df[self.depth_curve], 
            self.surface_temperature,
            self.gradient
        )
        
        
def rw_temp_convert(rw,t1,t2, temp_unit='f'):
    """rw_temp_convert [summary]

    Parameters
    ----------
    rw : [type]
        [description]
    t1 : [type]
        [description]
    t2 : [type]
        [description]
    temp_unit : str, optional
        [description], by default 'f'

    Returns
    -------
    [type]
        [description]
    """
    rw = np.atleast_1d(rw)
    t1 = np.atleast_1d(t1)
    t2 = np.atleast_1d(t2)
    
    if temp_unit=='f':
        c = np.array([6.77])
    else:
        c = np.array([21.5])
    
    rw2 = rw*((t1 + c)/(t2 + c))
    return rw2

class RwConverter(PetrophysicsEstimator):
    rw: float = Field(...)
    temperature: float = Field(...)
    temp_unit: str = Field('f')
    col_name: str = 'rw_temp'
    temp_curve: str = Field(None)
    
    def forward(self,to_temp):
        return rw_temp_convert(
            self.rw,
            self.temperature,
            to_temp,
            temp_unit=self.temp_unit
        )
        
def rw_salinity(temp, salinity,temp_unit='f'):
    """rw [summary]

    Parameters
    ----------
    temp : [type]
        [description]
    salinity : [type]
        [description]
    temp_unit : str, optional
        [description], by default 'f'

    Returns
    -------
    [type]
        [description]
    """
    
    # 1) Convert from Celcius to Farenheit
    if temp_unit=='c':
        tf = 1.8*temp + 32.0
    else:
        tf=temp
    
    # 2) Calculate Resistivity in Ohm meters
    rw = np.power((400000.0/(tf*salinity)),0.88)
    
    return rw

class RwSalinity(PetrophysicsEstimator):
    temperature: float = Field(..., gt=0)
    temp_unit: str = Field('f')
    col_name: str = 'rw_salinity'
    salinity_curve: str = Field(None)
    
    def forward(self,salinity):
        return rw_salinity(
            self.temperature,
            salinity,
            temp_unit=self.temp_unit
        )
        


def facies_dnp(rho_curve, ntr_curve,pef_curve,**kw):
    """facies_dnp [summary]

    Parameters
    ----------
    rho_curve : [type]
        [description]
    ntr_curve : [type]
        [description]
    pef_curve : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    rho_curve = np.atleast_1d(rho_curve)
    ntr_curve = np.atleast_1d(ntr_curve)
    pef_curve = np.atleast_1d(pef_curve)
    phi_rho_curve = phi_rho(rho_curve,**kw)
    phia_curve = phia(phi_rho_curve,ntr_curve)
    
    u = pef_curve*((rho_curve + 0.1833)/1.07)
    uma = (u - 0.398 * phia_curve)/(1-phia_curve)
    dga = (rho_curve - phia_curve)/(1-phia_curve)
    return uma, dga

def perm_timur(phie,swir,dperm=4.4,eperm=2.0,fluid='oil'):
    if fluid=="oil":
        Cperm=3400
    elif fluid=="gas":
        Cperm=340
    k=Cperm * np.power(phie,dperm) / np.power(swir,eperm)
    return k

class PermTimur(PetrophysicsEstimator):
    dperm: float = Field(4.4)
    eperm: float = Field(2.0)
    fluid: str = Field('oil')
    col_name: str = 'perm_timur'
    phie_curve: str = Field(None)
    swir_curve: float = Field(None)
    
    def forward(self,phie,swir=None):
        if swir is None:
            swir = self.swir_curve
        return perm_timur(
            phie,
            swir,
            dperm=self.dperm,
            eperm = self.eperm,
            fluid=self.fluid
        )

    def forward_log(self,df):
        return perm_timur(
            df[self.phie_curve],
            df[self.swir_curve],
            dperm=self.dperm,
            eperm = self.eperm,
            fluid=self.fluid
        )

def perm_morris(phie_curve,swir,dperm=6.0,eperm=2.0,fluid='oil'):
    if fluid=="oil":
        Cperm=65000
    elif fluid=="gas":
        Cperm=6500
    k=Cperm * np.power(phie,dperm) / np.power(swir,dperm) 
    return k

class PermMorris(PetrophysicsEstimator):
    dperm: float = Field(6.0)
    eperm: float = Field(2.0)
    fluid: str = Field('oil')
    col_name: str = 'perm_morris'
    phie_curve: str = Field(None)
    swir_curve: float = Field(None)

    def forward(self,phie,swir=None):
        if swir is None:
            swir = self.swir_curve
        return perm_morris(
            phie,
            swir,
            dperm=self.dperm,
            eperm = self.eperm,
            fluid=self.fluid
        )

    def forward_log(self,df):
        return perm_morris(
            df[self.phie_curve],
            df[self.swir_curve],
            dperm=self.dperm,
            eperm = self.eperm,
            fluid=self.fluid
        )
        
def perm_coates(phie, swir):
    k=(10.0*phie)**4 *((1.0-swir)/swir)**2
    return k

class PermCoates(PetrophysicsEstimator):
    col_name: str = 'perm_coates'
    phie_curve: str = Field(None)
    swir_curve: float = Field(None)
    
    def forward(self,phie,swir=None):
        if swir is None:
            swir = self.swir_curve
        return perm_coates(phie, swir)
    
    def forward_log(self,df):
        return perm_coates(
            df[self.phie_curve], 
            df[self.phie_curve]
        )

def rw2(T, Wse, verbose=False,Celcius=False):    
    """rw2 [    Uses method from textbook "Petrophysics" by Djebbar and Donaldson.
    
    Supposedly more accurate than the approach used in calc_Rw because Hx and 
    RwT account for non-linearlity in resistivity as a function of salinity.]

    Parameters
    ----------
    T : [type]
        [description]
    Wse : [type]
        [description]
    verbose : bool, optional
        [description], by default False
    Celcius : bool, optional
        [description], by default False

    Returns
    -------
    [type]
        [description]
    """

    
    #  Calculations:
    
    # 1) Convert from Celcius to Farenheit
    if Celcius==True:
        Tf = 9.0/5.0*T + 32.0
    else:
        Tf=T
    
    # 2) Calculate reference water resistivity @ 75 Degrees Farenheit
    Rw75 = 1.0/(2.74*10**-4 * Wse**0.955) + 0.0123
    
    # 3) Calculate nonlinear correction factors
    Xh = 10.0**(-0.3404*np.log10(Rw75) + 0.6414)
    
    # 4) Calculate Water Resistivity at Temperature T1.  Output is Ohm-m
    Rw = Rw75 * (75.0 + Xh)/(Tf + Xh)
       
    return Rw

def flow_capacity(height,perm_curve,pay_curve):
    """flow_capacity [summary]

    Parameters
    ----------
    height : [type]
        [description]
    perm_curve : [type]
        [description]
    pay_curve : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    perm_curve=np.nan_to_num(np.atleast_1d(perm_curve))
    pay_curve=np.nan_to_num(np.atleast_1d(pay_curve))
    height=np.atleast_1d(height)
    
    kh=height*perm_curve*pay_curve
    khcum=np.cumsum(kh)
    kht=np.sum(kh)
    khnorm=1-(khcum/kht)
    return kh, khnorm


class FlowCapacity(PetrophysicsEstimator):
    depth_curve: str = Field(None)
    perm_curve: str = Field(None)
    pay_curve: str = Field(None)   
    
    def forward(self,depth,perm_curve,pay_curve):
        h = np.gradient(depth)
        return flow_capacity(h,perm_curve,pay_curve)  

    def forward_log(self,df):
        h = np.gradient(df[self.depth_curve])
        return flow_capacity(
            h,
            df[self.perm_curve],
            df[self.pay_curve]
        )
            
    
      
"""
https://www.spec2000.net/14-swtdt.htm
PHIe = effective porosity (fractional)
SIGMA = TDT capture cross section log reading (capture units)
SIGMAM = capture cross section matrix value (capture units)
SIGW = capture cross section for water (capture units)
SIGHY = capture cross section for hydrocarbons (capture units)
SIGSH = capture cross section for shale (capture units)
SWtdt = water saturation from TDT (fractional)
Vsh = shale volume (fractional)
WS = water salinity (ppm NaCl)
"""
def sw_pnn(phie,vsh, sigma,sighy,sigsh,ws=None,sigw=None,sigmam=None):
    """sw_pnn [summary]

    Parameters
    ----------
    phie : [type]
        [description]
    vsh : [type]
        [description]
    sigma : [type]
        [description]
    sighy : [type]
        [description]
    sigsh : [type]
        [description]
    ws : [type], optional
        [description], by default None
    sigw : [type], optional
        [description], by default None
    sigmam : [type], optional
        [description], by default None

    Returns
    -------
    [type]
        [description]
    """

    phie=np.atleast_1d(phie)
    vsh=np.atleast_1d(vsh)
    sigma=np.atleast_1d(sigma)
    sighy=np.atleast_1d(sighy)
    sigsh=np.atleast_1d(sigsh)

    if sigw is None:
        sigw = 22 + 0.000404*ws 
    if sigmam is None:
        sigmam = (sigma - phie*sigw)/(1-phie)

    _a = sigma - sigmam 
    _b = phie*(sighy - sigmam)
    _c = vsh*(sigsh - sigmam)
    _d = phie*(sigw - sighy)
    sw = (_a - _b - _c)/(_d)
    sw[phie<=0] = 1.0

    return sw

def rw_from_sp(rmf=None, rmf_temp=75, res_temp=None, ssp=None,temp_unit='f', rw_75=False):
    """rw_from_sp [summary]

    Parameters
    ----------
    rmf : [type], optional
        [description], by default None
    rmf_temp : int, optional
        [description], by default 75
    res_temp : [type], optional
        [description], by default None
    ssp : [type], optional
        [description], by default None
    temp_unit : str, optional
        [description], by default 'f'
    rw_75 : bool, optional
        [description], by default False

    Returns
    -------
    [type]
        [description]
    """

    #https://www.spec2000.net/05-7rwsp.htm
    # 1) Convert from Celcius to Farenheit
    if temp_unit=='c':
        rmf_temp = 1.8*rmf_temp + 32.0
        res_temp = 1.8*res_temp + 32.0

    # Convert rmf @ rmf_temp to res_temp
    rmf_res = rw_temp_convert(rmf,rmf_temp,res_temp, temp_unit='f')

    #Estimate Mud filtrate equivalent resistivity
    if rmf_res > 0.1:
        rmfe = rmf_res*0.85
    else:
        rmfe = (146*rmf_res-5)/(337*rmf_res + 77)
    #Ksp
    ksp = 60 +0.122*res_temp

    #RSP
    rsp = np.power(10,-ssp/ksp)
    # Rwe. Equivalent Water Resistivity
    rwe = rmfe / rsp

    # Estimate Water resistivity from Equivalent water resistivity
    if rwe > 0.12:
        rw = np.power(10,0.69*rwe-0.24) - 0.58
    else:
        rw = (77*rwe + 5)/(146 - 337*rwe)

    if rw_75:
        rw = rw_temp_convert(rw,res_temp,75, temp_unit='f')

    return rw

def salinity_from_rw(rw=None, temp=75,temp_unit='f'):
    """salinity_from_rw [summary]

    Parameters
    ----------
    rw : [type], optional
        [description], by default None
    temp : int, optional
        [description], by default 75
    temp_unit : str, optional
        [description], by default 'f'

    Returns
    -------
    [type]
        [description]
    """
    if temp_unit=='c':
        temp = 1.8*temp + 32.0

    # Convert rmf @ rmf_temp to res_temp
    if temp != 75:
        rw_75 = rw_temp_convert(rw,temp,75, temp_unit='f')
    else:
        rw_75 = rw 

    exp = (3.562-np.log10(rw_75-0.0123))/(0.955)
    sal = np.power(10,exp)

    return sal



    





import matplotlib.pyplot as plt 
import matplotlib as mpl
import pandas as pd
import numpy as np
from typing import List,Union, Tuple, Dict
from pydantic import validate_arguments
from .utils import formations_plot, perforations_plot, correlations_plot


@validate_arguments(config=dict(arbitrary_types_allowed=True))
def caltrack(
    df: pd.DataFrame,
    cali: Union[str,list] = None, 
    bit: Union[str,list] = None, 
    ylims: Tuple[float,float] =None,
    cal_lims:Tuple[float,float]=(5,20),
    dtick:bool=False,
    fill:bool=False,
    fontsize: int=8,
    grid_numbers : Tuple[float,float] = (11,51),
    steps: Tuple[float,float]  = None,
    ax=None,
    formations: pd.DataFrame = None,
    units: pd.DataFrame = None,
    perforations: pd.DataFrame = None,
    correlations: pd.DataFrame = None,
    formations_kw:Dict={},
    units_kw:Dict={},
    perforations_kw:Dict={},
    correlations_kw:Dict={},
    cal_kw:Dict={},
    bit_kw:Dict={},
    depth_ref:str='md',
    cal_colormap: str='winter',
    bit_colormap: str='bone'
):
    list_axes = []
    cal=ax or plt.gca()
    
    def_cal_kw = {
    'color': 'black',
    'linestyle':'-',
    'linewidth': 1
    }
    
    for (k,v) in def_cal_kw.items():
        if k not in cal_kw:
            cal_kw[k]=v

    def_bit_kw = {
    'color': 'darkred',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_bit_kw.items():
        if k not in bit_kw:
            bit_kw[k]=v
            
    #Set ylims
    if ylims==None: #Depth Limits
        ylims=[df.index.min(),df.index.max()]

    cal.set_ylim([ylims[1],ylims[0]])
        
    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[0])
        minor_grid = np.linspace(ylims[0],ylims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(ylims[0],ylims[1],steps[0])
        minor_grid = np.arange(ylims[0],ylims[1],steps[1])
    
    depth = df.index if depth_ref=='md' else df[depth_ref] 

    if cali is not None:
        if isinstance(cali,str):
            cal.plot(df[cali],depth,**cal_kw)   #Plotting
        elif isinstance(cali,list):
            cmap = mpl.cm.get_cmap(cal_colormap,len(cal))
            for i,c in enumerate(cali):
                cal_kw['color']=cmap(i)
                cal.plot(df[c],depth,**cal_kw)
        
    if bit is not None:
        if isinstance(bit,str):
            cal.plot(df[bit],depth,**cal_kw)   #Plotting
        elif isinstance(cali,list):
            cmap = mpl.cm.get_cmap(bit_colormap,len(cal))
            for i,c in enumerate(bit):
                cal_kw['color']=cmap(i)
                cal.plot(df[c],depth,**cal_kw)
    

    cal.set_xlim(cal_lims)
    cal.set_xlabel("Caliper [in]")
    cal.set_xticks(np.linspace(cal_lims[0],cal_lims[1],4))
    cal.set_xticklabels(np.round(np.linspace(cal_lims[0],cal_lims[1],4),decimals=1))
    cal.xaxis.tick_top()
    cal.xaxis.set_label_position("top")
    cal.tick_params("both",labelsize=fontsize)
    cal.set_yticks(mayor_grid)
    cal.set_yticks(minor_grid,minor=True)       
    if dtick==True:
        cal.set_yticklabels(mayor_grid)
    else:
        cal.set_yticklabels([])
        
    if fill==True:
        cal.fill_betweenx(depth,df[cali],df[bit],where=(df[cali] > df[bit]),color="orange")
        cal.fill_betweenx(depth,df[cali],df[bit],where=(df[cali]< df[bit]),color="gray")
        
    #Add formations tops
    if formations is not None:
        cal = formations_plot(
            formations,
            depth_ref,
            ylims,
            cal_lims,
            cal,
            config=formations_kw
        )

    # #Add units tops
    if units is not None:
        cal = formations_plot(
            units,
            depth_ref,
            ylims,
            cal_lims,
            cal,
            config=units_kw
        )

    #  #Add Interval Perforated
    if perforations is not None:
        cal = perforations_plot(
            perforations,
            depth_ref,
            ylims,
            cal_lims,
            cal,
            config=perforations_kw
        )
        
    if correlations is not None:
        cal = correlations_plot(
            correlations,
            depth_ref,
            ylims,
            cal_lims,
            cal,
            config=correlations_kw
        )
        
    list_axes.append(cal)
    return list_axes


import numpy as np
import pandas as pd 
import matplotlib.pyplot as plt 
from scipy.interpolate import interp1d

def cutoff(log,phi=None,sw=None,vsh=None,elbow=0.9,swaq=0.7,ax=None, plot=True):
    """cutoff [summary]

    Parameters
    ----------
    log : [type]
        [description]
    phi : [type], optional
        [description], by default None
    sw : [type], optional
        [description], by default None
    vsh : [type], optional
        [description], by default None
    elbow : float, optional
        [description], by default 0.9
    swaq : float, optional
        [description], by default 0.7
    ax : [type], optional
        [description], by default None
    plot : bool, optional
        [description], by default True

    Returns
    -------
    [type]
        [description]
    """
    step=np.mean(np.diff(log.index))
    Hc=log.loc[log[sw]<swaq,phi]*(1-log.loc[log[sw]<swaq,sw])*step
    Hct=sum(Hc[~np.isnan(Hc)])
    vshr=np.linspace(0,1,20)
    phir=np.linspace(0,0.4,20)
    swr=np.linspace(0,1,20)
    
    ranges=[phir,swr,vshr]
    symbols=[phi,sw,vsh]
    
    hcl=[]
    co={}
    re=[]
    for (i,s) in enumerate(symbols):
        d=[]
        for j in ranges[i]:
            if i!=0:
                x=np.sum(Hc[log[s]<=j])/Hct
            else:
                x=np.sum(Hc[log[s]>=j])/Hct
            d.append(x)
        hcl.append(d)
        f = interp1d(hcl[i],ranges[i])
        co[s]=f(elbow)
    re.append(co)
    if plot:
        ax = ax or plt.gca()
        ax.plot(phir,hcl[0],color='red',label='Phie CutOff={}'.format(np.round(co[symbols[0]],decimals=2)))
        ax.hlines(elbow,0,1,linestyle='--')
        ax.vlines(co[symbols[0]],0,1,linestyle='--')
        ax.set_xlim([0,1])
        ax.set_ylim([0,1])
        ax.plot(swr,hcl[1],color='blue',label="Sw CutOff={}".format(np.round(co[symbols[1]],decimals=2)))
        ax.vlines(co[symbols[1]],0,1,linestyle='--')
        ax.plot(vshr,hcl[2],color='gold',label='Vsh CutOff={}'.format(np.round(co[symbols[2]],decimals=2)))       
        ax.vlines(co[symbols[2]],0,1,linestyle='--')
        ax.set_xlabel('Phi-Sw-Vsh CutOff') 
        ax.set_ylabel('Norm Hydrocarbon Column')
        ax.set_title('CutOff Estimation HC')
        ax.legend()
        re.append(ax)
    return re

def pickett(rw=0.15,a=1,m=2,n=2,swr=np.linspace(0.1,1,5),ax=None):
    pax= ax or plt.gca()
    Phi_rt1=np.power(rw,1/m)  #Phi at Rt=1. Log(phi)=-mlog(Rt/Rw)
    rts=(a*rw)/(np.power(Phi_rt1,m)*np.power(swr,n))
    Rti=rts*np.power(Phi_rt1,m)/a
    phir=np.logspace(-2,0,num=9)
    RT=[]
    for i in Rti:
        rtr=i*a*np.power(phir,-m)
        RT.append(rtr)
    
    for (k,i) in enumerate(RT):
        ax.loglog(i,phir,label='Sw={}'.format(swr[k]))
    pax.set_xticks(np.logspace(-1,3,5))
    pax.set_xticklabels(np.logspace(-1,3,5))
    pax.set_yticks(np.logspace(-2,0,3))
    pax.set_yticklabels(np.logspace(-2,0,3))
    pax.legend()
    pax.set_ylabel('Porosity[]')
    pax.set_xlabel('Resistivity [Ohm m]')
    pax.set_title('Pickett Plot- a={},m={},n={},rw={}'.format(a,m,n,rw))
    return pax

def windland(phirange=None,r35range=None, ax=None):
    wax= ax or plt.gca()
    
    if phirange is None:
        phirange=np.linspace(0.05,0.30,5)
    else:
        phirange=np.asarray(phirange)
    if r35range is None:
        r35range=np.logspace(0,1,5)
    else:
        r35range=np.asarray(r35range)
    
    A=0.735
    B=0.588
    C=0.8641
    K=np.zeros((phirange.shape[0],r35range.shape[0]))
    for (i,v) in enumerate(r35range):
        k=np.power((v*np.power(phirange*100,C)*np.exp(-A)),1/B)
        K[:,i]=k
        
    wax=plt.gca()
    wax.plot(phirange,K)
    wax.set_yscale('log')
    wax.set_xlabel('Phi[]')
    wax.set_ylabel('Permeability [md]')
    wax.legend([i for i in r35range])
    return wax



from pydantic import BaseModel, Field, validator, PrivateAttr, validate_arguments
from typing import List, Dict, Union, Optional, Tuple
import numpy as np
import pandas as pd
import geopandas as gpd
from scipy.interpolate import interp1d
from enum import Enum
import folium
from folium.plugins import MarkerCluster, MeasureControl,MousePosition
import matplotlib.pyplot as plt
from plotly import graph_objects as go
from shapely.geometry import LineString
#Local imports
from .mincurve import min_curve_method, vtk_survey
from .interpolate import interpolate_deviation, interpolate_position
from .projection import projection_1d

class InterpolatorsEnum(str,Enum):
    md_to_tvd = 'md_to_tvd'
    md_to_tvdss = 'md_to_tvdss'
    tvd_to_md = 'tvd_to_md'
    tvd_to_tvdss = 'tvd_to_tvdss'
    tvdss_to_md = 'tvdss_to_md'
    tvdss_to_tvd = 'tvdss_to_tvd'

class DepthRefsEnum(str,Enum):
    md = 'md'
    tvd = 'tvd'
    tvdss = 'tvdss'
    
    
class Survey(BaseModel):
    md: Optional[Union[List[float], np.ndarray]] = Field(None)
    inc: Optional[Union[List[float], np.ndarray]] = Field(None)
    azi: Optional[Union[List[float], np.ndarray]] = Field(None)
    tvd: Optional[Union[List[float], np.ndarray]] = Field(None)
    tvdss: Optional[Union[List[float], np.ndarray]] = Field(None)
    north_offset: Optional[Union[List[float], np.ndarray]] = Field(None)
    east_offset: Optional[Union[List[float], np.ndarray]] = Field(None)
    northing: Optional[Union[List[float], np.ndarray]] = Field(None)
    easting: Optional[Union[List[float], np.ndarray]] = Field(None)
    dleg: Optional[Union[List[float], np.ndarray]] = Field(None)
    fields: Dict[str,List[Union[float,str]]] = Field(None)
    crs: int = Field(None, gt=0) 
    _interpolators: Dict[InterpolatorsEnum,interp1d] = PrivateAttr({})
    
    def __repr__(self) -> str:
        return f"<Survey: {'Empty' if self.md is None else len(self.md)}>"
        
    
    
    @validator('md')
    def check_array_pressures_order(cls, v):
        v_array = np.array(v)
        diff = np.diff(v_array)
        if not np.all(diff>0):
            raise ValueError('Md must be increasing order')
        if not np.all(v_array>=0):
            raise ValueError('Md must be postive')
        return v
    
    class Config:
        extra = 'forbid'
        validate_assignment = True
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        json_encoders = {np.ndarray: lambda x: x.tolist()}
        
    def df(
        self,
        projection:bool=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        to_crs:int=None,
        horizontal_distance:bool=False
    ):
        surv_dict = self.dict()
        fields = surv_dict.pop('fields',None)
        
        if fields is not None:
            surv_dict.update(fields)
            
        _gdf = gpd.GeoDataFrame(
            surv_dict, 
            geometry=gpd.points_from_xy(surv_dict['easting'],surv_dict['northing'])
        )
        _gdf.crs = self.crs
        
        if to_crs:
            _gdf = _gdf.to_crs(to_crs)

        if projection == True:
            _pr,c = projection_1d(_gdf[['easting','northing']].values, azi, center=center)
            _gdf['projection'] = _pr
            
        if horizontal_distance == True:
            shifted_geo = _gdf[['geometry']].shift(periods=1)
            _gdf['h_distance'] = _gdf['geometry'].distance(shifted_geo).fillna(0)
            _gdf['h_distance_cum'] = _gdf['h_distance'].cumsum()
        return _gdf
    
    def plot_projection(
        self,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        to_crs:int=None,
        ax=None
    ):
        df = self.df(projection=True,azi=azi,center=center,to_crs=to_crs)
        sax = ax or plt.gca()
        sax.plot(df['projection'],df['tvdss'])
        return sax
        
    
    def make_vertical(
        self,
        td,
        surface_northing=0,
        surface_easting=0, 
        kbe = 0,
        n=2
    ):
        self.md = np.linspace(0,td,n)
        self.inc = np.zeros(n)
        self.azi = np.zeros(n)
        self.tvd = np.linspace(0,td,n)
        self.tvdss = (self.tvd - kbe)*-1
        self.north_offset = np.zeros(n)
        self.east_offset = np.zeros(n)
        self.northing = np.full(n,surface_northing)
        self.easting =  np.full(n,surface_easting)
        self.dleg = np.zeros(n)
        
        return self
        
        
    
    def from_deviation(
        self,
        df:pd.DataFrame=None, 
        md:Union[str,list,np.ndarray, pd.Series]='md',
        azi:Union[str,list,np.ndarray, pd.Series]='azi',
        inc:Union[str,list,np.ndarray, pd.Series]='inc',
        surface_northing:float=0,
        surface_easting:float=0, 
        kbe:float=0, 
        crs=None
    ):
        
        if crs is None:
            crs = self.crs
        else:
            self.crs = crs
        
        dev_dict = {}
        for i,name in zip([md,azi,inc],['md','azi','inc']):
            if isinstance(i,str):
                _var = df[i].values
            else:
                _var = np.atleast_1d(i)
            dev_dict[name] = _var
            
        gdf = min_curve_method(
            dev_dict['md'],
            dev_dict['inc'],
            dev_dict['azi'],
            surface_northing=surface_northing,
            surface_easting=surface_easting,
            kbe=kbe,
            crs = crs
        )
        
        for j in ['md','inc','azi','tvd','tvdss','north_offset','east_offset','northing','easting','dleg']:
            setattr(self,j,gdf[j].tolist())
        
        #? Is it good practice return self??
        return self
    
    def horizontal_length(self):
        return LineString(self.df()['geometry'].tolist()).length
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def sample_deviation(
        self,
        step:Optional[float]=100, 
        min_depth:Optional[float]=None, 
        max_depth:Optional[float]=None,
        depth_ref:str='md',
        depth_array:Optional[np.ndarray]=None
    )->pd.DataFrame:

        _survey = self.df()
        
        if min_depth is not None:
            _survey = _survey.loc[_survey[depth_ref]>=min_depth]

        if max_depth is not None:
            _survey = _survey.loc[_survey[depth_ref]<=max_depth] 
                
        return interpolate_deviation(
            _survey['md'].values, 
            _survey['inc'].values, 
            _survey['azi'].values, 
            md_step=step,
            md_array=depth_array
        )           

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def sample_survey(
        self,          
        step:Optional[float]=100, 
        min_depth:Optional[float]=None, 
        max_depth:Optional[float]=None,
        depth_ref:str='md',
        depth_array:Optional[np.ndarray]=None
    ):
        
        new_dev = self.sample_deviation(
            step = step,
            min_depth = min_depth,
            max_depth = max_depth,
            depth_ref = depth_ref,
            depth_array=depth_array
        )
        
        if min_depth is not None:
            coord = self.coord_interpolate(
                values=new_dev['new_md'].iloc[0],
                depth_ref='md'
            )
            
            easting = coord['geometry'].iloc[0].x
            northing = coord['geometry'].iloc[0].y
        else:
            easting = self.easting[0]
            northing = self.northing[0]
        
        return Survey().from_deviation(
            new_dev,
            md = 'new_md',
            inc = 'new_inc',
            azi = 'new_azi',
            surface_easting=easting,
            surface_northing=northing,
            crs=self.crs
        )
        
    
    @validate_arguments
    def sample_depth(
        self,          
        step:float=100, 
        min_depth:Optional[float]=None, 
        max_depth:Optional[float]=None,
        depth_ref:str='md'
    ):
        
        _survey = self.df()
        
        if min_depth is not None:
            _survey = _survey.loc[_survey[depth_ref]>=min_depth]

        if max_depth is not None:
            _survey = _survey.loc[_survey[depth_ref]<=max_depth]
            
        new_pos = interpolate_position(
            _survey['tvd'], 
            _survey['easting'], 
            _survey['northing'], 
            tvd_step=step
        )
        
        new_pos['new_md'] = self.depth_interpolate(
            values=new_pos['new_tvd'].values,
            mode = 'tvd_to_md'    
        )
        
        
        
        return gpd.GeoDataFrame(
            new_pos,
            geometry=gpd.points_from_xy(new_pos.new_easting,new_pos.new_northing),
            crs=self.crs
        )
    
    @validate_arguments
    def sample_position(
        self,
        step:float = 100,
        
    ):
        surv = self.df(
            horizontal_distance=True
        )
        
        sample = np.arange(
            surv['h_distance_cum'].min(),
            surv['h_distance_cum'].max()+step,
            step
        )
        sample[-1] = surv['h_distance_cum'].max()
        
        inter = interp1d(
            surv['h_distance_cum'],
            surv['md'],
            fill_value='extrapolate'
        )(sample)
        
        return self.sample_survey(
            step=None,
            depth_array=inter
        )
        
        
        
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def depth_interpolate(
        self,
        values:Union[float,List[float],np.ndarray]=None,
        mode:InterpolatorsEnum=InterpolatorsEnum.md_to_tvd
    ):
        if self._interpolators is None or mode.value not in self._interpolators:
            str_split = mode.value.split('_')
            _survey=self.df()
            self._interpolators[mode.value] = interp1d(_survey[str_split[0]],_survey[str_split[2]],fill_value='extrapolate')

        if values is None:
            return self._interpolators[mode.value](self.df()['md'])
        else:
            return self._interpolators[mode.value](np.atleast_1d(values))     
    
    @validate_arguments
    def coord_interpolate(
        self,
        values:Union[float,List[float]]=None,
        depth_ref:DepthRefsEnum=DepthRefsEnum.md,
    ):

        if depth_ref != DepthRefsEnum.tvd:
            values_tvd = self.depth_interpolate(values = values, mode = f'{depth_ref.value}_to_tvd')
        else:
           
            values_tvd = np.atleast_1d(values)

        if any([self._interpolators is None,'tvd_easting' not in self._interpolators]):
            _survey=self.df()
            self._interpolators['tvd_easting'] = interp1d(_survey['tvd'],_survey['easting'],fill_value='extrapolate')
            self._interpolators['tvd_northing'] = interp1d(_survey['tvd'],_survey['northing'],fill_value='extrapolate')

        if values is None:
            easting = self._interpolators['tvd_easting'](self.df()['tvd'])
            northing = self._interpolators['tvd_northing'](self.df()['tvd'])
        else:

            easting = self._interpolators['tvd_easting'](values_tvd)
            northing = self._interpolators['tvd_northing'](values_tvd)
        
                
        return gpd.GeoDataFrame(
            geometry=gpd.points_from_xy(easting,northing),
            crs = self.crs
        )
        
    def get_vtk(self, to_crs = None):
    
        _survey = self.df()
        
        if to_crs:
            _survey = _survey.to_crs(to_crs)
            
        _survey['x'] = _survey['geometry'].x
        _survey['y'] = _survey['geometry'].y
            
        surv_vtk = vtk_survey(_survey[['x','y','tvdss']].values)
        
        for col in _survey.iteritems():
            surv_vtk.point_data[col[0]] = col[1].values

        return surv_vtk
    
    def survey_map(
        self,
        zoom:int=10,
        map_style:str = 'OpenStreetMap',
        tooltip:bool=True,
        popup:bool=False,
        ax=None,
        radius=10
    ):

        _coord = self.df()
        _coord = _coord.to_crs('EPSG:4326')
        _coord['lon'] = _coord['geometry'].x
        _coord['lat'] = _coord['geometry'].y
        center = _coord[['lat','lon']].mean(axis=0)
        
        #make the map
        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            assert isinstance(ax,folium.folium.Map)
            map_folium = ax
            
        for i, r in _coord.iterrows():
            folium.Circle(
                [r['lat'],r['lon']],
                tooltip=f"md:{r['md']} <br>tvd:{r['tvd']} <br>tvdss:{r['tvdss']} <br>inc:{r['inc']} " if tooltip else None,
                popup = folium.Popup(html=f"<br>md:{r['md']} <br>tvd:{r['tvd']} <br>tvdss:{r['tvdss']} <br>inc:{r['inc']} ",show=True,max_width='50%') if popup else None,
                #icon=folium.Icon(icon='circle',prefix='fa', color='green'),
                radius=radius
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    def trace3d(
        self, 
        area_unit='m', 
        depth_unit='ft',
        name='survey',
        mode = 'lines',
        **kwargs
    ):
        
        df = self.df()
        
        coef_area = 1 if area_unit == 'm' else 3.28084
        coef_depth = 1 if depth_unit == 'ft' else 0.3048
        
        trace = go.Scatter3d(
            x = df['easting'] * coef_area,
            y = df['northing'] * coef_area,
            z = df['tvdss'] * coef_depth,
            hoverinfo = 'all',
            mode = mode,
            name = name,
            **kwargs
        )
        
        return trace
    
    def plot3d(
        self, 
        area_unit='m', 
        depth_unit='ft',
        name='survey',
        mode = 'lines',
        title = 'Survey',
        trace_kw = {},
        width = 800,
        height = 600,
        **kwargs
    ):
        
        traces = self.trace3d(
            area_unit=area_unit,
            depth_unit=depth_unit,
            name = name,
            mode = mode,
            **trace_kw
        )
        
        layout = go.Layout(
            title = title,
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **kwargs
        )
        
        return go.Figure(data=[traces],layout=layout)
            




## Code taken from https://github.com/Zabamund/wellpathpy was modified.


import numpy as np
import geopandas as gpd
import pyvista as pv
#local imports
from .checkarrays import checkarrays


# class Survey(gpd.GeoDataFrame):

#     def __init__(self, *args, **kwargs): 
#         md = kwargs.pop("md", None)                                                                                                                                  
#         super(Survey, self).__init__(*args, **kwargs)

#         if md is not None:
#             md = np.atleast_1d(md)
#             self['md'] = md
#             assert self['md'].is_monotonic, "md must be increasing"
#             self.set_index('md',inplace=True)
#             self.index.name='md'
#         elif 'md' in self.columns:
#             assert self['md'].is_monotonic, "md must be increasing"
#             self.set_index('md',inplace=True)
#             self.index.name='md'
#         elif self.index.name == 'md':
#             assert self.index.is_monotonic, "md must be increasing"
    
   
#     @property
#     def _constructor(self):
#         return Survey

def minimum_curvature(md, inc, azi):
    """Minimum curvature

    Calculate TVD, northing, easting, and dogleg, using the minimum curvature
    method.

    This is the inner workhorse of the min_curve_method, and only implement the
    pure mathematics. As a user, you should probably use the min_curve_method
    function.

    This function considers md unitless, and assumes inc and azi are in
    radians.

    Parameters
    ----------
    md : array_like of float
        measured depth
    inc : array_like of float
        inclination in radians
    azi : array_like of float
        azimuth in radians

    Returns
    -------
    tvd : array_like of float
        true vertical depth
    northing : array_like of float
    easting : array_like of float
    dogleg : array_like of float

    Notes
    -----
    This function does not insert surface location
    """
    md, inc, azi = checkarrays(md, inc, azi)


    # extract upper and lower survey stations
    md_upper, md_lower = md[:-1], md[1:]
    inc_upper, inc_lower = inc[:-1], inc[1:]
    azi_upper, azi_lower = azi[:-1], azi[1:]

    cos_inc = np.cos(inc_lower - inc_upper)
    sin_inc = np.sin(inc_upper) * np.sin(inc_lower)
    cos_azi = 1 - np.cos(azi_lower - azi_upper)

    dogleg = np.arccos(cos_inc - (sin_inc * cos_azi))

    # ratio factor, correct for dogleg == 0 values
    rf = 2 / dogleg * np.tan(dogleg / 2)
    rf = np.where(dogleg == 0., 1, rf)

    md_diff = md_lower - md_upper

    upper = np.sin(inc_upper) * np.cos(azi_upper)
    lower = np.sin(inc_lower) * np.cos(azi_lower) * rf
    northing = np.cumsum(md_diff / 2 * (upper + lower))

    upper = np.sin(inc_upper) * np.sin(azi_upper)
    lower = np.sin(inc_lower) * np.sin(azi_lower) * rf
    easting = np.cumsum(md_diff / 2 * (upper + lower))

    tvd = np.cumsum(md_diff / 2 * (np.cos(inc_upper) + np.cos(inc_lower)) * rf)

    return tvd, northing, easting, dogleg

def min_curve_method(md, inc, azi, md_units='ft', norm_opt=0,surface_northing=0,surface_easting=0, kbe=0, crs=None):
    """
    Calculate TVD using minimum curvature method.

    This method uses angles from upper and lower end of survey interval to
    calculate a curve that passes through both survey points. This curve is
    smoothed by use of the ratio factor defined by the tortuosity or dogleg
    of the wellpath.

    Formula
    -------
    dls = arccos(cos(inc_lower - inc_upper) - sin(inc_upper) * sin(inc_lower) * (1 - cos(azi_lower - azi_upper)))
    rf = 2 / dls * (tan(dls/2))
    northing = sum((md_lower - md_upper) * (sin(inc_upper) * cos(azi_upper) + sin(inc_lower) * cos(azi_lower) / 2) * cf)
    easting = sum((md_lower - md_upper) *(sin(inc_upper) * sin(azi_upper) + sin(inc_lower) * sin(azi_lower) / 2) * cf)
    tvd = sum((md_lower - md_upper) * (cos(inc_lower) + cos(inc_upper) / 2) * cf)

    where:
    dls: dog leg severity (degrees)
    rf: ratio factor (radians)
    md_upper: upper survey station depth MD
    md_lower: lower survey station depth MD
    inc_upper: upper survey station inclination in degrees
    inc_lower: lower survey station inclination in degrees
    azi_upper: upper survey station azimuth in degrees
    azi_lower: lower survey station azimuth in degrees

    Parameters
    ----------
    md: float, measured depth in m or ft
    inc: float, well deviation in degrees
    azi: float, well azimuth in degrees
    md_units: str, measured depth units in m or ft
        used for dogleg severity calculation
    norm_opt: float, dogleg normalisation value,
        if passed will override md_units

    Returns
    -------
    Deviation converted to TVD, easting, northing
        tvd in m,
        northing in m,
        easting in m
    Dogleg severity
        dls: dogleg severity angle in degrees per normalisation value
            (normalisation value is deg/100ft, deg/30m or deg/<norm_opt>)

    Notes
    -----
    tvd are in ft, northing and easting in meters.
    The user must convert to feet if required.

    """

    # get units and normalising for dls
    try:
        norm_opt + 0
    except TypeError:
        raise TypeError('norm_opt must be a float')

    if norm_opt != 0:
        norm = norm_opt
    else:
        if md_units == 'm':
            norm = 30
        elif md_units == 'ft':
            norm = 100
        else:
            raise ValueError('md_units must be either m or ft')

    md, inc, azi = checkarrays(md, inc, azi)
    inc_rad = np.deg2rad(inc)
    azi_rad = np.deg2rad(azi)

    md_diff = md[1:] - md[:-1]
    tvd, northing, easting, dogleg = minimum_curvature(md, inc_rad, azi_rad)

    tvd = np.insert(tvd, 0, 0)
    northing_off = np.insert(northing, 0, 0)
    easting_off = np.insert(easting, 0, 0)

    # calculate dogleg severity, change md units if dls in ft is passed in
    dl = np.rad2deg(dogleg)
    dls = dl * (norm / md_diff)
    dls = np.insert(dls, 0, 0)


    northing = surface_northing + (northing_off*0.3048)
    easting = surface_easting + (easting_off*0.3048)
    tvdss = (tvd - kbe)*-1

    survdf = gpd.GeoDataFrame({'md':md,'inc':inc,'azi':azi,'tvd':tvd,'tvdss':tvdss,
                            'north_offset':northing_off,'east_offset':easting_off,
                            'northing':northing,'easting':easting,'dleg':dls}, 
                             geometry=gpd.points_from_xy(easting,northing)
                             )
    
    if crs is not None:
        assert isinstance(crs,(int,str))
        survdf.crs = crs
  
    return survdf


def vtk_survey(points:np.ndarray):
    """vtk_survey [    Transforms a x,y,z numpy array into a pyvista.PolyData object 
                    that represents a well survey]

    Parameters
    ----------
    points : np.ndarray
        [A 2D numpy array with a values of (x,y,z) coordinates of a well]

    Returns
    -------
    [pyvista.PolyData]
        [Pyvista object spline that represents the well survey]
    """

    poly = pv.PolyData()
    poly.points = points
    cells = np.full((len(points)-1, 3), 2, dtype=np.int)
    cells[:, 1] = np.arange(0, len(points)-1, dtype=np.int)
    cells[:, 2] = np.arange(1, len(points), dtype=np.int)
    poly.lines = cells
    return poly


## Code taken from https://github.com/Zabamund/wellpathpy


import numpy as np
from pydantic import validate_arguments
from scipy import interpolate
import pandas as pd
from .checkarrays import checkarrays, checkarrays_tvd, checkarrays_monotonic_tvd


def interpolate_deviation(
    md, 
    inc, 
    azi, 
    md_step=None, 
    md_array=None
):
    """
    Interpolate a well deviation to a given step.

    Parameters
    ----------
    md: float, measured depth (units not defined)
    inc: float, well inclination in degrees from vertical
    azi: float, well azimuth in degrees from North
    md_step: int or float, md increment to interpolate to

    Returns
    -------
    Deviation intepolated to new md_step:
        md, inc, azi

    Notes
    -----
    This function should not be used before md->tvd conversion.
    Note that the input arrays must not contain NaN values.

    """

    md, inc, azi = checkarrays(md, inc, azi)

    for input_array in [md, inc, azi]:
        if np.isnan(input_array).any():
            raise ValueError('md, inc and azi cannot contain NaN values.')

    try:
        if md_step is not None:
            new_md = np.arange(md.min(), md.max() + md_step, md_step)
            new_md[-1] = md.max()
        else:
            new_md = md_array
        
    except TypeError:
        raise TypeError('md_step must be int or float')

    f_inc = interpolate.interp1d(md, inc)
    new_inc = f_inc(new_md)
    f_azi = interpolate.interp1d(md, azi)
    new_azi = f_azi(new_md)

    new_deviation = pd.DataFrame({'new_md':new_md,'new_inc':new_inc,'new_azi':new_azi})
    return new_deviation

def interpolate_position(tvd, easting, northing, tvd_step=1):
    """
    Interpolate a well positional log to a given step.

    Parameters
    ----------
    tvd: float, true verical depth (units not defined)
    northing: float, north-offset from zero reference point
        the units should be the same as the input deviation
        or the results will be wrong
    easting: float, east-offset from zero reference point
        the units should be the same as the input deviation
        or the results will be wrong
    tvd_step: int or float, tvd increment to interpolate to

    Returns
    -------
    Deviation intepolated to new step:
        tvd, easting, northing

    Notes
    -----
    This function should not be used before tvd->md conversion.
    Note that the input arrays must not contain NaN values.
    The tvd values must be strictly increasing, i.e. this
    method will not work on horizontal wells, use
    `interpolate_deviation` for those wells.

    """
    tvd, easting, northing = checkarrays_monotonic_tvd(tvd, easting, northing)

    for input_array in [tvd, northing, easting]:
        if np.isnan(input_array).any():
            raise ValueError('tvd, northing and easting cannot contain NaN values.')

    try:
        new_tvd = np.arange(tvd[0], tvd[-1] + tvd_step, tvd_step)
        new_tvd[-1] = tvd[-1]
    except TypeError:
        raise TypeError('tvd_step must be int or float')

    f_easting = interpolate.interp1d(tvd, easting)
    new_easting = f_easting(new_tvd)
    f_northing = interpolate.interp1d(tvd, northing)
    new_northing = f_northing(new_tvd)

    new_position = pd.DataFrame({'new_tvd':new_tvd,'new_easting':new_easting,'new_northing':new_northing})
    return new_position


## Code taken from https://github.com/Zabamund/wellpathpy


import numpy as np

def checkarrays(md, inc, azi):
    """
    Assure basic preconditions are met, and convert input (md, inc, azi) to
    numpy arrays.

    This function will ensure that:

    - All inputs are convertible to arrays-of-floats, and perform this
      conversion
    - All inputs are of the same shape
    - md is strictly increasing

    Parameters
    ----------
    md: array_like of float
        measured depth
    inc: array_like of float
        well deviation
    azi: array_like of float
        azimuth

    Returns
    -------
    md: array_like of float
        measured depth
    inc: array_like of float
        well deviation
    azi: array_like of float
        azimuth

    Raises
    ------
    ValueError
        If md, inc, or azi, are of different shapes
        If the md values are not strictly increasing
    """
    md = np.asarray(md, dtype = np.float)
    inc = np.asarray(inc, dtype = np.float)
    azi = np.asarray(azi, dtype = np.float)

    if not (md.shape == inc.shape == azi.shape):
        raise ValueError('md, inc, and azi must be the same shape')

    if not np.all(md[1:] > md[:-1]):
        raise ValueError('md must have strictly increasing values')

    return md, inc, azi

def checkarrays_tvd(tvd, northing, easting):
    """
    Assure basic preconditions are met, and convert input (tvd, northing, easting) to
    numpy arrays.

    This function will ensure that:

    - All inputs are convertible to arrays-of-floats, and perform this
      conversion
    - All inputs are of the same shape

    Parameters
    ----------
    tvd: array_like of float
        true vertical depth
    northing: array_like of float
        north-offset
    easting: array_like of float
        east-offset

    Returns
    -------
    tvd: array_like of float
        true vertical depth
    northing: array_like of float
        north-offset
    easting: array_like of float
        east-offset

    Raises
    ------
    ValueError
        If tvd, northing, or easting, are of different shapes
    """
    tvd = np.asarray(tvd, dtype = np.float)
    northing = np.asarray(northing, dtype = np.float)
    easting = np.asarray(easting, dtype = np.float)

    if not (tvd.shape == northing.shape == easting.shape):
        raise ValueError('tvd, northing, and easting must be the same shape')

    return tvd, northing, easting

def checkarrays_monotonic_tvd(tvd, northing, easting):
    """
    Assure basic preconditions are met, and convert input (tvd, northing, easting) to
    numpy arrays.

    This function will ensure that:

    - All inputs are convertible to arrays-of-floats, and perform this
      conversion
    - All inputs are of the same shape
    - tvd is strictly increasing

    Parameters
    ----------
    tvd: array_like of float
        true vertical depth
    northing: array_like of float
        north-offset
    easting: array_like of float
        east-offset

    Returns
    -------
    tvd: array_like of float
        true vertical depth
    northing: array_like of float
        north-offset
    easting: array_like of float
        east-offset

    Raises
    ------
    ValueError
        If tvd, northing, or easting, are of different shapes
        If the tvd values are not strictly increasing
    """
    tvd = np.asarray(tvd, dtype = np.float)
    northing = np.asarray(northing, dtype = np.float)
    easting = np.asarray(easting, dtype = np.float)

    if not (tvd.shape == northing.shape == easting.shape):
        raise ValueError('tvd, northing, and easting must be the same shape')
    if not np.all(tvd[1:] > tvd[:-1]):
        raise ValueError('tvd must have strictly increasing values')

    return tvd, northing, easting


from fieldspy.wells.projection import projection_1d
from pydantic import BaseModel, Field, validate_arguments, validator, parse_obj_as
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import geopandas as gpd
from typing import Tuple, Dict, List, Optional, Union
from shapely.geometry import Point
from lasio import LASFile
from wellschematicspy import WellSchema
import folium
from folium.plugins import MeasureControl, MousePosition
from scipy.spatial import distance_matrix, Delaunay
import seaborn as sns
import pyvista as pv
from sqlalchemy.engine.base import Engine
from plotly import graph_objects as go
from datetime import date
#local imports
from .survey import Survey
from .perforations import Perforations, Perforation
from .formations import  Formations, Formation
from .units import Units, Unit
from .tools import Tool, Tools
from .log import Las

# Folium Colors available
folium_color_list = ['red', 'blue', 'green', 'purple', 'orange', 'darkred',
    'lightred', 'beige', 'darkblue', 'darkgreen', 'cadetblue', 'darkpurple', 
    'white', 'pink', 'lightblue', 'lightgreen', 'gray', 'black', 'lightgray']


class Well(BaseModel):
    name: str = Field(...)
    surface_coordinate: Union[Tuple[float,float],Point, Dict[str,float]] = Field(None)
    crs: int = Field(None)
    rte: float = Field(0.0)
    survey: Survey = Field(None)
    logs: Dict[str,Las] = Field(None)
    diagram: Dict[str,WellSchema] = Field(None)
    formations: Formations = Field(None)
    units: Units = Field(None)
    tools: Tools = Field(None)
    perforations: Perforations = Field(None)
    platform: str = Field(None)
    field: str = Field(None)
    block: str = Field(None)
    basin: str = Field(None)
    spud_date: date = Field(None)
    completion_date: date = Field(None)
    td_md: float = Field(None)
    td_tvd: float = Field(None)
    fields: Dict[str,Union[float,int,str]] = Field(None)
    
    class Config:
        extra = 'ignore'
        validate_assignment = True
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        json_encoders = {
            Point: lambda x: {'x': x.x, 'y': x.y},   
            LASFile: lambda x: x.to_json(),
            np.ndarray: lambda x: x.tolist()
        }

    def __str__(self):
        return self.well_head().T.dropna().astype(str).reset_index().to_string()

    def __repr__(self) -> str:
        return self.__str__()
    
    @validator('surface_coordinate',always=True)
    def get_surf_coord(cls, v):
        if isinstance(v,tuple):
            return Point(v[0],v[1])
        elif isinstance(v,dict):
            return Point(v['x'],v['y'])
        return v

    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def survey_from_deviation(self,df:pd.DataFrame, md:str=None,inc:str=None,azi:str=None):
        surv = Survey(crs=self.crs)
        surv.from_deviation(
            df=df,
            md=md,
            inc=inc, 
            azi=azi,
            surface_northing=self.surface_coordinate.y,
            surface_easting=self.surface_coordinate.x
        )
        self.survey = surv
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def add_logs(self, logs: Dict[str,Las]):
        if self.logs:
            self.logs.update(logs)
        self.logs = logs

    @validate_arguments
    def add_diagram(self, diagrams: Dict[str,WellSchema]):
        if self.diagram:
            self.diagram.update(diagrams)
        self.diagram = diagrams
    
    @validate_arguments
    def add_formations(self, formations: Union[Formation,List[Formation],Dict[str,Formation]]):
        if self.formations:
            self.formations.add_formations(formations)
        self.formations = Formations(formations=formations)
        
    @validate_arguments
    def add_units(self, units: Union[Unit,List[Unit],Dict[str,Unit]]):
        if self.units:
            self.units.add_units(units)
        self.units = Units(units=units)
        
    @validate_arguments
    def add_tools(self, tools: Union[Tool,List[Tool],Dict[str,Tool]]):
        if self.tools:
            self.tools.add_tools(tools)
        self.tools = Tools(tools=tools)
        
    @validate_arguments
    def add_perforations(self, perforations: Union[Formation,List[Formation],Dict[str,Formation]]):
        if self.perforations:
            self.perforations.add_perforations(perforations)
        self.perforations = Perforations(perforations=perforations)
    
    @validate_arguments
    def well_map(self,zoom:int=10, map_style:str = 'OpenStreetMap', to_crs:int=4326, tooltip:bool=False,popup:bool=True, ax=None):
        """
        Make a Foluim map with the selected well

        Input:
            zoom -> (int, float) Initial zoom for folium map
            map_stule -> (str) Type of map folium
        Return:
            w_map -> (folium.Map) Folium map object
        """
        _coord = gpd.GeoDataFrame()

        x_coord = self.surface_coordinate.x
        y_coord = self.surface_coordinate.y
        geometry = self.surface_coordinate
        crs = self.crs
        _w = gpd.GeoDataFrame({'x':[x_coord],'y':[y_coord],'geometry':[geometry]}, index=[self.name])
        _w.crs = crs
        _w = _w.to_crs(to_crs)
        _w['lon'] = _w['geometry'].x
        _w['lat'] = _w['geometry'].y
        _coord = _coord.append(_w)
        center = _coord[['lat','lon']].mean(axis=0)

        #make the map
        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            assert isinstance(ax,folium.folium.Map)
            map_folium = ax

        for i, r in _coord.iterrows():
            folium.Marker(
                [r['lat'],r['lon']],
                tooltip=f"{i}" if tooltip else None,
                popup = folium.Popup(html=f"{i}",show=True) if popup else None,
                icon=folium.Icon(icon='tint', color='green')
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    def describe(self):
        dict_attr = {
            'surface_coordinate': True if self.surface_coordinate else False,
            'crs': True if self.crs else False,
            'rte': True if self.rte else False,
            'survey': True if self.survey else False,
            'logs': True if self.logs else False,
            'diagram': True if self.diagram else False,
            'formations': True if self.formations else False,
            'units': True if self.units else False,
            'tools': True if self.tools else False,
            'perforations': True if self.perforations else False,
            'platform': True if self.platform else False,      
            'field': True if self.field else False,
            'block': True if self.block else False,
            'basin': True if self.basin else False,
            'spud_date': True if self.spud_date else False,
            'completion_date': True if self.completion_date else False,
            'td_md': True if self.td_md else False,
            'td_tvd': True if self.td_tvd else False         
        }
        return pd.Series(dict_attr, name=self.name)
    
    def well_head(self,to_crs:int=None,c:float=0.3028):
        if to_crs is None:
            to_crs = self.crs
        _w = gpd.GeoDataFrame({
            'field': [self.field],
            'block': [self.block],
            'basin': [self.basin],
            'z':[self.rte * c],
            'geometry':[self.surface_coordinate],
            'platform':[self.platform],
            'spud_date':[self.spud_date],
            'completion_date':[self.completion_date],
            'td_md':[self.td_md],
            'td_tvd':[self.td_tvd]            
        }, index=[self.name])
        
        cols_fields = []
        if self.fields is not None:
            for f in self.fields:
                _w[f] = [self.fields[f]]
                cols_fields.append(f)
                
        _w.crs = self.crs
        
        _w = _w.to_crs(to_crs)

        _w['x'] = _w['geometry'].x
        _w['y'] = _w['geometry'].y
        _w['crs'] = to_crs
        
        #converto to wsg
        _wsg = _w.to_crs(4326)
        _w['lon'] = _wsg['geometry'].x
        _w['lat'] = _wsg['geometry'].y
        
        cols = ['field','block','basin','x','y','z','crs','geometry','lon','lat','platform','spud_date','completion_date','td_md','td_tvd']
        _w.index.name='well'
        _w = _w.round(1)
        return _w[cols + cols_fields]

    
class Wells(BaseModel):
    name: Optional[str] = Field(None)
    wells: Dict[str,Well] = Field(None)
    
    class Config:
        extra = 'ignore'
        validate_assignment = True
        underscore_attrs_are_private = True
        arbitrary_types_allowed = True
        json_encoders = {
            Point: lambda x: {'x': x.x, 'y': x.y},   
            LASFile: lambda x: x.to_json(),
            np.ndarray: lambda x: x.tolist()
        }
    
    def __str__(self):
        return f'Wells Name: {self.name if self.name is not None else "Name not set"}\nWells: {len(self.wells) if self.wells is not None else "Empty"}'
    
    def __repr__(self) -> str:
        return self.__str__()
    
    def __getitem__(self, key: str) -> Well:
        return self.wells[key]
    
    def __len__(self):
        return 0 if self.wells is None else len(self.wells)
    
    def __getitem__(self, key):
        return self.wells[key]
    
    def __iter__(self):
        return (self.wells[w] for w in self.wells)
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def wells_filter(
        self,
        wells: Optional[Union[str,List[str]]] = None,
        field: Optional[Dict] = None
    ):
        
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
            
        if field is None:
            return list_wells
        
        list_wells_filtered_full = []
        #for loop fields filter
        for f in field:
            # for loop wells allowed
            list_wells_filtered = []
            for w in list_wells:
                #if field in filter not in well object continue
                if f not in self.wells[w].fields:
                    continue
                # if field filter is a list check if well object field is in list
                if isinstance(field[f],list):
                    if self.wells[w].fields[f] not in field[f]:
                        continue
                # if field filter is a dict apply operators lt, le, gt, ge, eq, ne
                elif isinstance(field[f],dict):
                    if 'lt' in field[f]:
                        if self.wells[w].fields[f] < field[f]['lt']:
                            pass
                        else:
                            continue
                    if 'le' in field[f]:
                        if self.wells[w].fields[f] <= field[f]['le']:
                            pass
                        else:
                            continue
                    if 'gt' in field[f]:
                        if self.wells[w].fields[f] > field[f]['gt']:
                            pass
                        else:
                            continue
                    if 'ge' in field[f]:
                        if self.wells[w].fields[f] >= field[f]['ge']:
                            pass
                        else:
                            continue
                    if 'eq' in field[f]:
                        if self.wells[w].fields[f] == field[f]['eq']:
                            pass
                        else:
                            continue
                    if 'ne' in field[f]:
                        if self.wells[w].fields[f] != field[f]['ne']:
                            pass
                        else:
                            continue
                else:
                    if self.wells[w].fields[f] != field[f]:
                        continue
                    
                list_wells_filtered.append(w)
            list_wells_filtered_full.append(list_wells_filtered)
        
        return list(set.intersection(*map(set, list_wells_filtered_full)))         

    def add_wells(self, wells: Dict[str,Well]):
        if self.logs:
            self.logs.update(wells)
        self.logs = wells
        
    def describe(self):
        list_series = []
        for well in self.wells:
            list_series.append(self.wells[well].describe())
        return pd.concat(list_series, axis=1).T
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_well_heads(self,df:pd.DataFrame,name:str=None,fields=None,**kwargs):
        df = df.copy()
               
        if name:
            df['name'] = df[name].copy()
            df = df.set_index(name)
        else:
            df['name'] = df.index
            
        #To change columns name to match Object
        if bool(kwargs):
            kwargs = {v: k for k, v in kwargs.items()}
            df = df.rename(columns=kwargs)
            
        if fields is not None:
            fields_dict = df[fields].to_dict(orient='index')
            df.drop(fields,axis=1,inplace=True)
            
            wells_dict = df.to_dict(orient='index')
            
            for well in fields_dict:
                wells_dict[well].update({'fields':fields_dict[well]})
        else:
            wells_dict = df.to_dict(orient='index')
        
        
        self.wells = parse_obj_as(Dict[str,Well],wells_dict)

    
    @validate_arguments
    def formations(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        formations:Optional[Union[str,List[str]]]=None,
        projection:Optional[bool]=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        estimate_midpoint:bool=False,
        midpoint_weights_column:Optional[str]='md_tick',
        midpoint_name:str='midpoint'
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_formations_list = []
        for well in list_wells:
            if self.wells[well].formations is None:
                continue
            gf = self.wells[well].formations
            #Estimate Mid point
            if estimate_midpoint:
                try:
                    gf.estimate_average_midpoint(
                        formations=formations,
                        weights=midpoint_weights_column,
                        name = midpoint_name
                    )
                except:
                    pass            

            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_formations_list.append(gdf)
        
        if len(gdf_formations_list) == 0:
            raise ValueError('No formations found')
        gdf_formations = gpd.GeoDataFrame(
            pd.concat(gdf_formations_list),
            crs=gdf_formations_list[0].crs
        )

        if formations:
            gdf_formations = gdf_formations.loc[formations]
        
        _pr = None
        if projection:
            _pr,c = projection_1d(
                gdf_formations[['easting','northing']].values, 
                azi, 
                center=center
            )
            gdf_formations['projection'] = _pr
                
        return gdf_formations
    
    @validate_arguments
    def units(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        units:Optional[Union[str,List[str]]]=None,
        projection:Optional[bool]=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        estimate_midpoint:bool=False,
        midpoint_weights_column:Optional[str]='md_tick',
        midpoint_name:str='midpoint'
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_units_list = []
        for well in list_wells:
            if self.wells[well].units is None:
                continue
            gf = self.wells[well].units
            
            #Estimate Mid point
            if estimate_midpoint:
                try:
                    gf.estimate_average_midpoint(
                        units=units,
                        weights=midpoint_weights_column,
                        name = midpoint_name
                    )
                except:
                    pass
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_units_list.append(gdf)

        if len(gdf_units_list) == 0:
            raise ValueError('No units found')
        gdf_units = gpd.GeoDataFrame(
            pd.concat(gdf_units_list),
            crs=gdf_units_list[0].crs
        )

        if units:
            gdf_units = gdf_units.loc[units]
        
        _pr = None
        if projection:
            _pr,c = projection_1d(
                gdf_units[['easting','northing']].values, 
                azi, 
                center=center
            )
            gdf_units['projection'] = _pr
                
        return gdf_units
    
    @validate_arguments
    def tools(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        tools:Optional[Union[str,List[str]]]=None,
        projection:Optional[bool]=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        estimate_midpoint:bool=False,
        midpoint_weights_column:Optional[str]='md_tick',
        midpoint_name:str='midpoint'
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_tools = gpd.GeoDataFrame()
        for well in list_wells:
            if self.wells[well].tools is None:
                continue
            gf = self.wells[well].tools
            
            #Estimate Mid point
            if estimate_midpoint:
                try:
                    gf.estimate_average_midpoint(
                        tools=tools,
                        weights=midpoint_weights_column,
                        name = midpoint_name
                    )
                except:
                    pass
                
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_tools = gdf_tools.append(gdf)

        if tools:
            gdf_tools = gdf_tools.loc[tools]
        
        _pr = None
        if projection:
            _pr,c = projection_1d(
                gdf_tools[['easting','northing']].values, 
                azi, 
                center=center
            )
            gdf_tools['projection'] = _pr
                
        return gdf_tools
    
    @validate_arguments
    def formations_projection(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        formations:Optional[Union[str,List[str]]]=None,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_formations = gpd.GeoDataFrame()
        for well in list_wells:
            if self.wells[well].formations is None:
                continue            
            gf = self.wells[well].formations
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_formations = gdf_formations.append(gdf)

        if formations:
            gdf_formations = gdf_formations.loc[formations]
            
        _pr,c = projection_1d(
            gdf_formations[['easting','northing']].values, 
            azi, 
            center=center
        )
        gdf_formations['projection'] = _pr
                
        return gdf_formations, c
    
    @validate_arguments
    def units_projection(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        units:Optional[Union[str,List[str]]]=None,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_units = gpd.GeoDataFrame()
        for well in list_wells:
            if self.wells[well].units is None:
                continue            
            gf = self.wells[well].units
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_units = gdf_units.append(gdf)

        if units:
            gdf_units = gdf_units.loc[units]
            
        _pr,c = projection_1d(
            gdf_units[['easting','northing']].values, 
            azi, 
            center=center
        )
        gdf_units['projection'] = _pr
                
        return gdf_units, c
    
    @validate_arguments
    def perforations_projection(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        perforations:Optional[Union[str,List[str]]]=None,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_perforations = gpd.GeoDataFrame()
        for well in list_wells:
            if self.wells[well].perforations is None:
                continue            
            gf = self.wells[well].perforations
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_perforations = gdf_perforations.append(gdf)

        if perforations:
            gdf_perforations = gdf_perforations.loc[perforations]
            
        _pr,c = projection_1d(
            gdf_perforations[['easting','northing']].values, 
            azi, 
            center=center
        )
        gdf_perforations['projection'] = _pr
                
        return gdf_perforations, c
    
    @validate_arguments
    def tools_projection(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        tools:Optional[Union[str,List[str]]]=None,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_tools = gpd.GeoDataFrame()
        for well in list_wells:
            if self.wells[well].tools is None:
                continue            
            gf = self.wells[well].tools
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_tools = gdf_tools.append(gdf)

        if tools:
            gdf_tools = gdf_tools.loc[tools]
            
        _pr,c = projection_1d(
            gdf_tools[['easting','northing']].values, 
            azi, 
            center=center
        )
        gdf_tools['projection'] = _pr
                
        return gdf_tools, c
    
    @validate_arguments
    def surveys(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        projection:Optional[bool]=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        to_crs:int=None
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
        
        surveys_gdf = gpd.GeoDataFrame()
        gdf_surveys_list = []
        for well in list_wells:
            if self.wells[well].survey is None:
                continue
            
            surv_gdf = self.wells[well].survey.df()
            if to_crs:
                surv_gdf = surv_gdf.to_crs(to_crs)
            surv_gdf['well'] = self.wells[well].name
            gdf_surveys_list.append(surv_gdf)

        surveys_gdf = gpd.GeoDataFrame(
            pd.concat(gdf_surveys_list),
            crs=gdf_surveys_list[0].crs
        )
            
        if projection == True:
            _pr,c = projection_1d(surveys_gdf[['easting','northing']].values, azi, center=center)
            surveys_gdf['projection'] = _pr
            
        return surveys_gdf

    @validate_arguments
    def surveys_ascii(
        self, 
        wells:Union[str,List[str]]=None, 
        factor=None, 
        cols:List[str]=['easting','northing','tvdss','md'],
        float_format='{:.2f}'.format
    ):
               
        wells_surveys_df = self.surveys(wells=wells)
             
        string = ""

        if factor is None:
            factor = np.ones(len(cols))
        else:
            factor = np.atleast_1d(factor)
            assert (factor.ndim==1) & (factor.shape[0]==len(cols))
        
        for w in wells_surveys_df['well'].unique():

            _df = wells_surveys_df.loc[wells_surveys_df['well']==w,cols] * factor
            string += f"WELLNAME: {w}\n"
            string += _df.to_string(header=False,index=False,float_format=float_format) + '\n'
        return string
    
    @validate_arguments
    def perforations(
        self, 
        wells:Optional[Union[str,List[str]]]=None,
        perforations:Optional[Union[str,List[str]]]=None,
        projection:Optional[bool]=False,
        azi: Optional[float]=90,
        center: Optional[Tuple[float,float]] = None,
        estimate_midpoint:bool=False,
        midpoint_weights_column:Optional[str]='md_tick',
        midpoint_name:str='midpoint'
    ):

        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
           
        gdf_perforations_list = []
        for well in list_wells:
            if self.wells[well].perforations is None:
                continue
            gf = self.wells[well].perforations
            
            #Estimate Mid point
            if estimate_midpoint:
                try:
                    gf.estimate_average_midpoint(
                        perforations=perforations,
                        weights=midpoint_weights_column,
                        name = midpoint_name
                    )
                except Exception as e:
                    print(e)
            
            if self.wells[well].survey:
                gf.get_reference_depth(self.wells[well].survey)
                gf.estimate_ticks()
                gf.estimate_midpoints()
                gf.estimate_coordinate(self.wells[well].survey)
            gdf = gf.df()
            gdf['well'] = self.wells[well].name
            gdf_perforations_list.append(gdf)

        if len(gdf_perforations_list) == 0:
            raise ValueError('No perforations found')
        gdf_perforations = gpd.GeoDataFrame(
            pd.concat(gdf_perforations_list),
            crs=gdf_perforations_list[0].crs
        )

        if perforations:
            gdf_perforations = gdf_perforations.loc[perforations]
        
        if projection:
            _pr,c = projection_1d(
                gdf_perforations[['easting','northing']].values, 
                azi, 
                center=center
            )
            gdf_perforations['projection'] = _pr
                
        return gdf_perforations
    
    @validate_arguments
    def perforations_ascii(self,
        wells:Optional[List[str]]=None, 
        perforations:Optional[List[str]]=None,
        factor:Optional[float]=None, 
        cols:List[str]=['md_top','md_bottom'],
        float_format='{:.2f}'.format
    ):
        
        wells_perforations_df = self.perforations(
            wells=wells, 
            perforations=perforations
        ).reset_index()
             
        string = ""

        if factor is None:
            factor = np.ones(len(cols))
        else:
            factor = np.atleast_1d(factor)
            assert (factor.ndim==1) & (factor.shape[0]==len(cols))
            
        wells_perforations_df['completion'] = 'perforation'
        
        if 'date' not in wells_perforations_df.columns:
            wells_perforations_df['date'] = '"SOH"'
        else:
            wells_perforations_df['date'] = wells_perforations_df['date'].apply(lambda x: x.strftime('%Y-%m-%d').upper())
        
        if 'skin' not in wells_perforations_df.columns:
            wells_perforations_df['skin'] = 0

        if 'OH' not in wells_perforations_df.columns:
            wells_perforations_df['oh'] = 0.354           
        
        for w in wells_perforations_df['well'].unique():
            #_df = wells_perforations_df.loc[wells_perforations_df['well']==w,:]
            wells_perforations_df.loc[wells_perforations_df['well']==w,cols] = wells_perforations_df.loc[wells_perforations_df['well']==w,cols] * factor        
            
            string += f"WELLNAME {w}\n"
            cols_order = ['date','completion','md_top','md_bottom','oh','skin']
            string += wells_perforations_df.loc[wells_perforations_df['well']==w,cols_order].to_string(header=False,index=False,float_format=float_format) + '\n'
        return string
    
    @validate_arguments
    def well_heads(
        self, 
        wells:Optional[List[str]]=None, 
        to_crs:Optional[int]=None,
        c:float=0.3028
    ):
        
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))

        #Create coordinates dataframe
        _coord = []
        
        for well in list_wells:
            _w = self.wells[well].well_head(to_crs=to_crs,c=c)
            _coord.append(_w)
        
        df = gpd.GeoDataFrame(
            pd.concat(_coord, axis=0),
            crs = _coord[0].crs
        )
        
        return df
        
    @validate_arguments
    def wells_coordinates(
        self, 
        wells:Optional[List[str]]=None, 
        to_crs:Optional[int]=4326,
        c:float=0.3028
    ):


        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))

        #Create coordinates dataframe
        _coord = gpd.GeoDataFrame()

        for well in list_wells:
            x_coord = self.wells[well].surface_coordinate.x
            y_coord = self.wells[well].surface_coordinate.y
            z_coord = self.wells[well].rte * c
            shape = self.wells[well].surface_coordinate
            crs = self.wells[well].crs
            _w = gpd.GeoDataFrame({'x':[x_coord],'y':[y_coord],'z':[z_coord],'geometry':[shape]}, index=[well])
            _w.crs = crs
            _w = _w.to_crs(to_crs)
            _w['lon'] = _w['geometry'].x
            _w['lat'] = _w['geometry'].y
            _coord = _coord.append(_w)

        return _coord
    
    @validate_arguments
    def wells_distance(
        self,
        wells:Optional[List[str]]=None,
        dims: List[str] = ['x','y','z'],
        to_crs:int=None,
        stack:bool=False,     
        lt:Optional[float]=None,
        gt:Optional[float]=0.,
        le:Optional[float]=None,
        ge:Optional[float]=None,
        eq:Optional[float]=None,
        ne:Optional[float]=None,
        between:Optional[Tuple[float,float]]=None
    ):
        
        coords = self.well_heads(wells=wells,to_crs=to_crs)

        dist_array = distance_matrix(coords[dims].values,coords[dims].values)
        dist_matrix = pd.DataFrame(dist_array,index=coords.index.values, columns=coords.index.values)
        dist_matrix.index.name = 'well_row'
        dist_matrix.columns.name = 'well_column'
        
        locs = locals()
        if any([locs[i] is not None for i in ['lt','gt','le','ge','eq','ne','between']]):
            dist_matrix_s = dist_matrix.stack().reset_index()
            dist_matrix_s.columns = ['well_row','well_column','distance']
            
            filter_list = []
            for f in ['lt','gt','le','ge','eq','ne','between']:
                if locs[f] is None:
                    continue
                if f == 'between':
                    filter = eval(f"dist_matrix_s['distance'].{f}(*{locs[f]})")
                else:
                    filter = eval(f"dist_matrix_s['distance'].{f}({locs[f]})")
                filter_list.append(filter)
                
            fil_df = pd.concat(filter_list, axis=1).product(axis=1).astype(bool)
            dist_matrix_sf = dist_matrix_s.loc[fil_df,:]
            
            if stack:
                return dist_matrix_sf
            else:
                return dist_matrix_sf.pivot(index='well_row',columns='well_column',values='distance')
            
        if stack:
            dist_matrix = dist_matrix.stack().reset_index()
            dist_matrix.columns = ['well_row','well_column','distance']
            return dist_matrix
        return dist_matrix
    

    
    @validate_arguments
    def formation_distance(
        self,
        formation:str,
        wells:Optional[List[str]]=None,
        dims:list=['x','y','z'],
        stack:bool=False,
        lt:Optional[float]=None,
        gt:Optional[float]=0.,
        le:Optional[float]=None,
        ge:Optional[float]=None,
        eq:Optional[float]=None,
        ne:Optional[float]=None,
        between:Optional[Tuple[float,float]]=None
    ):
        coords = self.formations(wells=wells,formations=formation)
        coords[['x','y']] = coords[['easting','northing']]
        coords['z'] = coords['tvdss_top']*0.3028

        dist_array = distance_matrix(coords[dims].values,coords[dims].values)
        dist_matrix = pd.DataFrame(dist_array,index=coords['well'], columns=coords['well'])
        dist_matrix.index.name = 'well_row'
        dist_matrix.columns.name = 'well_column'

        locs = locals()
        if any([locs[i] is not None for i in ['lt','gt','le','ge','eq','ne','between']]):
            dist_matrix_s = dist_matrix.stack().reset_index()
            dist_matrix_s.columns = ['well_row','well_column','distance']
            
            filter_list = []
            for f in ['lt','gt','le','ge','eq','ne','between']:
                if locs[f] is None:
                    continue
                if f == 'between':
                    filter = eval(f"dist_matrix_s['distance'].{f}(*{locs[f]})")
                else:
                    filter = eval(f"dist_matrix_s['distance'].{f}({locs[f]})")
                filter_list.append(filter)
                
            fil_df = pd.concat(filter_list, axis=1).product(axis=1).astype(bool)
            dist_matrix_sf = dist_matrix_s.loc[fil_df,:]
            
            if stack:
                return dist_matrix_sf
            else:
                return dist_matrix_sf.pivot(index='well_row',columns='well_column',values='distance')
            
        if stack:
            dist_matrix = dist_matrix.stack().reset_index()
            dist_matrix.columns = ['well_row','well_column','distance']
            return dist_matrix
        return dist_matrix
        
    
    @validate_arguments
    def wells_map(
      self,
      wells:Optional[List[str]]=None,
      zoom:int =10,
      map_style:str = 'OpenStreetMap',
      tooltip=True,
      popup=False,
      ax=None,
      crs:int = 4326,
      color:Union[str,Dict[str,str]] = 'green',
      hue:str = None,
      seed:int=21,
    ):
        
        coords = self.well_heads(wells=wells,to_crs = crs)
        center = coords[['lat','lon']].mean(axis=0)

        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            map_folium = ax
            
        if hue is None:
            coords['color'] = color
        else:
            if not isinstance(color,dict):
                unique_type = coords[hue].unique().tolist()
                rs = np.random.RandomState(seed)
                colors_type = rs.choice(folium_color_list,size=len(unique_type),replace=False)
                color = dict(zip(unique_type,colors_type))
            coords['color'] = coords[hue].map(color)
                

        for i, r in coords.iterrows():
            folium.Marker(
                [r['lat'],r['lon']],
                tooltip=f"{i} {r[hue] if hue is not None else ''}" if tooltip else None,
                popup = folium.Popup(html=f"{i} {r[hue] if hue is not None else ''}",show=True,max_width='50%') if popup else None,
                icon=folium.Icon(icon='tint', color=r['color'])
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    @validate_arguments
    def formations_map(
        self,
        wells:Optional[List[str]]=None,
        formations:Optional[List[str]]=None,
        zoom:int =10,
        map_style:str = 'OpenStreetMap',
        tooltip=True,
        popup=False,
        ax=None,
        crs:int = 4326, 
    ):
        
        coords = self.formations(wells=wells, formations=formations)
        coords = coords.to_crs(crs)
        coords['lon'] = coords['geometry'].x
        coords['lat'] = coords['geometry'].y
        center = coords[['lat','lon']].mean(axis=0)
        
        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            map_folium = ax
            
        for i, r in coords.iterrows():
            folium.Marker(
                [r['lat'],r['lon']],
                tooltip=f"{r['well']} {i}" if tooltip else None,
                popup = folium.Popup(html=f"{r['well']} {i}",show=True,max_width='50%') if popup else None,
                icon=folium.Icon(icon='tint', color='green')
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    @validate_arguments
    def units_map(
        self,
        wells:Optional[List[str]]=None,
        units:Optional[List[str]]=None,
        zoom:int =10,
        map_style:str = 'OpenStreetMap',
        tooltip=True,
        popup=False,
        ax=None,
        crs:int = 4326, 
    ):
        
        coords = self.units(wells=wells, units=units)
        coords = coords.to_crs(crs)
        coords['lon'] = coords['geometry'].x
        coords['lat'] = coords['geometry'].y
        center = coords[['lat','lon']].mean(axis=0)
        
        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            map_folium = ax
            
        for i, r in coords.iterrows():
            folium.Marker(
                [r['lat'],r['lon']],
                tooltip=f"{r['well']} {i}" if tooltip else None,
                popup = folium.Popup(html=f"{r['well']} {i}",show=True,max_width='50%') if popup else None,
                icon=folium.Icon(icon='tint', color='green')
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    @validate_arguments
    def surveys_map(
        self,
        wells:Optional[List[str]]=None,
        zoom:int =10,
        map_style:str = 'OpenStreetMap',
        tooltip=True,
        popup=False,
        ax=None,
        crs:int = 4326, 
        radius:int = 10,
    ):
        
        coords = self.surveys(wells=wells)
        coords = coords.to_crs(crs)
        coords['lon'] = coords['geometry'].x
        coords['lat'] = coords['geometry'].y
        center = coords[['lat','lon']].mean(axis=0)
        
        if ax is None:
            map_folium = folium.Map(
                location=(center['lat'],center['lon']),
                zoom_start=zoom,
                tiles = map_style)
        else:
            map_folium = ax
            
        for i, r in coords.iterrows():
            folium.Circle(
                [r['lat'],r['lon']],
                tooltip=f"{r['well']} <br>md:{r['md']} <br>tvd:{r['tvd']} <br>tvdss:{r['tvdss']} <br>inc:{r['inc']} " if tooltip else None,
                popup = folium.Popup(html=f"{r['well']} <br>md:{r['md']} <br>tvd:{r['tvd']} <br>tvdss:{r['tvdss']} <br>inc:{r['inc']} ",show=True,max_width='50%') if popup else None,
                #icon=folium.Icon(icon='circle',prefix='fa', color='green'),
                radius=radius
                ).add_to(map_folium)

        folium.LayerControl().add_to(map_folium)
        #LocateControl().add_to(map_folium)
        MeasureControl().add_to(map_folium)
        MousePosition().add_to(map_folium)

        return map_folium
    
    @validate_arguments
    def structural_view3d(
        self,
        wells:Optional[List[str]]=None,
        by:Optional[str] = None,
        formations:Optional[List[str]]=None,
        units:Optional[List[str]]=None,
        perforations:Optional[List[str]]=None,
        tools:Optional[List[str]]=None,
        show_surveys:bool=False,
        show_formations:bool=False,
        show_units:bool=False,
        show_perforations:bool=False,
        show_tools:bool=False,
        formations_scatter:bool=True,
        units_scatter:bool=True,
        perforations_scatter:bool=True,
        tools_scatter:bool=True,
        surveys_mode:str='lines+markers',
        width:int=800,
        height:int=800,
        perforations_triangulate:bool=False,
        formations_triangulate:bool=False,
        units_triangulate:bool=False,
        tools_triangulate:bool=False,
        layout_kw:Dict={}
    ): 
        list_traces = []
        if by is not None:
            list_of_list_of_wells = []
            wh_df =  self.well_heads(wells=wells)
            for j in wh_df[by].unique():
                list_of_list_of_wells.append(wh_df.loc[wh_df[by]==j].index.tolist())
            
            for list_of_wells in list_of_list_of_wells:
                if show_formations:
                    fm_traces = self.formations3d_traces(
                        wells=list_of_wells,
                        formations=formations,
                        scatter=formations_scatter,
                        triangulate=formations_triangulate
                    )
                    list_traces.extend(fm_traces)
                    
                if show_units:
                    un_traces = self.units3d_traces(
                        wells=list_of_wells,
                        units=units,
                        scatter=units_scatter,
                        triangulate=units_triangulate
                    )
                    list_traces.extend(un_traces)
                
                if show_perforations:
                    perf_traces = self.perforations3d_traces(
                        wells=list_of_wells,
                        perforations=perforations,
                        scatter=perforations_scatter,
                        triangulate = perforations_triangulate
                    )
                    list_traces.extend(perf_traces)
                    
                if show_tools:
                    perf_traces = self.tools3d_traces(
                        wells=list_of_wells,
                        tools=tools,
                        scatter=tools_scatter,
                        triangulate = tools_triangulate
                    )
                    list_traces.extend(perf_traces)   
        else:
            if show_formations:
                fm_traces = self.formations3d_traces(
                    wells=wells,
                    formations=formations,
                    scatter=formations_scatter,
                    triangulate=formations_triangulate
                )
                list_traces.extend(fm_traces)
                
            if show_units:
                un_traces = self.units3d_traces(
                    wells=wells,
                    units=units,
                    scatter=units_scatter,
                    triangulate=units_triangulate
                )
                list_traces.extend(un_traces)
            
            if show_perforations:
                perf_traces = self.perforations3d_traces(
                    wells=wells,
                    perforations=perforations,
                    scatter=perforations_scatter,
                    triangulate = perforations_triangulate
                )
                list_traces.extend(perf_traces)
                
            if show_tools:
                perf_traces = self.tools3d_traces(
                    wells=wells,
                    tools=tools,
                    scatter=tools_scatter,
                    triangulate = tools_triangulate
                )
                list_traces.extend(perf_traces)
        
        if show_surveys:
            surv_traces = self.surveys3d_traces(
                wells=wells,
                mode=surveys_mode
            )
            list_traces.extend(surv_traces)
            
        layout = go.Layout(
            title = 'Surveys',
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **layout_kw
        )
        return go.Figure(
            data=list_traces,
            layout=layout
        )
            
    @validate_arguments
    def structural_view(
        self,
        wells:Optional[List[str]]=None,
        formations:Optional[List[str]]=None,
        units:Optional[List[str]]=None,
        perforations:Optional[List[str]]=None,
        tools:Optional[List[str]]=None,
        show_surveys:bool=True,
        show_formations:bool=False,
        show_units:bool=False,
        show_perforations:bool=False,
        show_tools:bool=False,
        azi:float = 0,
        center:Optional[Tuple[float,float]]=None,
        margin:float=500,
        ax=None,
        fm_color:str='Set1',
        well_color:str='Set2',
        unit_color:str='magma',
        perforation_color:str='viridis',
        tools_color:str='crest',
        legend:Union[bool,str]='brief',
        formation_kind:str='line',
        unit_kind:str='line',
        perforation_kind:str='scatter',
        well_kind:str='scatter',
        tool_kind:str='scatter',
        ann_formations:bool=False,
        ann_units:bool=False,
        ann_perforations:bool=False,
        ann_tools:bool=False,
        ann_fontsize:float=11,
        ylims:Optional[Tuple[float,float]]=None,
        **kwargs
    ):
        stax= ax or plt.gca()
               
        if show_formations:
            tops, center = self.formations_projection(
                wells = wells,
                formations=formations,
                azi=azi,
                center=center
            )
            tops.reset_index(inplace=True)
            
            if formation_kind == 'line':
                sns.lineplot(
                    x='projection',
                    y='tvdss_top', 
                    data=tops, 
                    hue='formation',
                    markers=True, 
                    ax=stax, 
                    palette=fm_color, 
                    legend=legend
                )
            else:
                sns.scatterplot(
                    x='projection',
                    y='tvdss_top', 
                    data=tops, 
                    hue='formation',
                    markers=True, 
                    ax=stax, 
                    palette=fm_color, 
                    legend=legend)
                
            if ann_formations:
                for i,v in tops.iterrows():
                    stax.annotate(
                        f"{v['well']}",
                        xy=(v['projection'],v['tvdss_top']),
                        xycoords='data',
                        horizontalalignment='right', 
                        fontsize=ann_fontsize,
                        bbox={'boxstyle':'round', 'fc':'0.8'},
                        xytext=(0, 20),
                        textcoords='offset points'
                    )

        if show_tools:
            tools_tops, center = self.tools_projection(
                wells = wells,
                tools=tools,
                azi=azi,
                center=center
            )
            tools_tops.reset_index(inplace=True)
            
            if tool_kind == 'line':
                sns.lineplot(
                    x='projection',
                    y='tvdss_top', 
                    data=tools_tops, 
                    hue='tool',
                    markers=True, 
                    ax=stax, 
                    palette=tools_color, 
                    legend=legend
                )
            else:
                sns.scatterplot(
                    x='projection',
                    y='tvdss_top', 
                    data=tools_tops, 
                    hue='tool',
                    markers=True, 
                    ax=stax, 
                    palette=tools_color, 
                    legend=legend)
                
            if ann_tools:
                for i,v in tops.iterrows():
                    stax.annotate(
                        f"{v['well']}",
                        xy=(v['projection'],v['tvdss_top']),
                        xycoords='data',
                        horizontalalignment='right', 
                        fontsize=ann_fontsize,
                        bbox={'boxstyle':'round', 'fc':'0.8'},
                        xytext=(0, 20),
                        textcoords='offset points'
                    )
                
        if show_units:
            units_top, center = self.units_projection(
                wells = wells,
                units=units,
                azi=azi,
                center=center
            )
            units_top.reset_index(inplace=True)
            
            if unit_kind == 'line':
                sns.lineplot(
                    x='projection',
                    y='tvdss_top', 
                    data=units_top, 
                    hue='unit',
                    markers=True, 
                    ax=stax, 
                    palette=unit_color, 
                    legend=legend
                )
            else:
                sns.scatterplot(
                    x='projection',
                    y='tvdss_top', 
                    data=units_top, 
                    hue='unit',
                    markers=True, 
                    ax=stax, 
                    palette=unit_color, 
                    legend=legend
                )

            if ann_units:
                for i,v in units_top.iterrows():
                    stax.annotate(
                        f"{v['well']}",
                        xy=(v['projection'],v['tvdss_top']),
                        xycoords='data',
                        horizontalalignment='right', 
                        fontsize=ann_fontsize,
                        bbox={'boxstyle':'round', 'fc':'0.8'},
                        xytext=(0, 20),
                        textcoords='offset points'
                    )

        if show_perforations:
            perfs_top, center = self.perforations_projection(
                wells = wells,
                perforations=perforations,
                azi=azi,
                center=center
            )
            perfs_top.reset_index(inplace=True)
            
            if perforation_kind == 'line':
                sns.lineplot(
                    x='projection',
                    y='tvdss_top', 
                    data=perfs_top, 
                    hue='perforation',
                    markers=True, 
                    ax=stax, 
                    palette=perforation_color, 
                    legend=legend
                )
            else:
                sns.scatterplot(
                    x='projection',
                    y='tvdss_top', 
                    data=perfs_top, 
                    hue='perforation',
                    markers=True, 
                    ax=stax, 
                    palette=unit_color, 
                    legend=legend
                )
                
            if ann_perforations:
                for i,v in perfs_top.iterrows():
                    stax.annotate(
                        f"{v['well']}",
                        xy=(v['projection'],v['tvdss_top']),
                        xycoords='data',
                        horizontalalignment='right', 
                        fontsize=ann_fontsize,
                        bbox={'boxstyle':'round', 'fc':'0.8'},
                        xytext=(0, 20),
                        textcoords='offset points'
                    )

                
        if show_surveys:
            surv= self.surveys(
                wells=wells,
                projection=True, 
                azi=azi, 
                center=center
            ).reset_index(drop=True)
            
            if well_kind=='line':
                sns.lineplot(
                    x='projection',
                    y='tvdss', 
                    data=surv, 
                    hue='well', 
                    ax=stax, 
                    palette=well_color, 
                    legend=legend
                )
            else:
                sns.scatterplot(
                    x='projection',
                    y='tvdss', 
                    data=surv, 
                    hue='well', 
                    ax=stax, 
                    palette=well_color, 
                    legend=legend
                )

        if ylims==None: #Depth Limits
            if show_surveys and show_formations:
                ylims=[surv['tvdss'].max()-margin,surv['tvdss'].min()+margin]
            elif show_surveys:
                ylims=[surv['tvdss'].max()-margin,surv['tvdss'].min()+margin]
            elif show_formations:
                ylims=[tops['tvdss_top'].max()-margin,tops['tvdss_top'].min()+margin]
            elif show_perforations:
                ylims=[perfs_top['tvdss_top'].max()-margin,perfs_top['tvdss_top'].min()+margin]

        stax.set_ylim([ylims[1],ylims[0]])

        xlims = kwargs.pop('xlims',None)
        if xlims:
            stax.set_xlim([xlims[0],xlims[1]])
            
        return stax
    
    @validate_arguments
    def surveys_vtk(
        self,
        wells: Optional[List[str]] = None,
        to_crs:int = None,
        add_wellhead:bool = False
    ):
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
            
        survey_blocks = pv.MultiBlock()
        for well in list_wells:
            survey_block = self.wells[well].survey.get_vtk(to_crs=to_crs)
            if add_wellhead:
                wh = self.wells[well].well_head(to_crs=to_crs)
                for i in wh.columns:
                    if wh[i].iloc[0] is not None and i != 'geometry':
                        survey_block.add_field_data([wh[i].iloc[0]],i)
            survey_blocks[well] = survey_block
        
        return survey_blocks
    
    @validate_arguments
    def surveys3d_traces(
        self,
        wells: Optional[List[str]] = None,
        area_unit:str='m', 
        depth_unit:str='ft',
        mode:str = 'lines',
        trace_kw:Dict = {}
    ):
        
        list_wells = []
        if isinstance(wells,str):
            list_wells.append(wells)
        elif isinstance(wells,list):
            list_wells.extend(wells)
        else:
            list_wells.extend(list(self.wells.keys()))
            
        traces = []
        for well in list_wells:
            if self.wells[well].survey is None:
                continue
            trace = self.wells[well].survey.trace3d(name=well, area_unit=area_unit, depth_unit=depth_unit, mode = mode, **trace_kw)
            traces.append(trace)        
        
        return traces

    @validate_arguments
    def surveys3d(
        self,
        wells: Optional[List[str]] = None,
        area_unit:str='m', 
        depth_unit:str='ft',
        mode:str = 'lines',
        width:float = 700,
        height:float = 700,
        trace_kw:Dict = {},
        layout_kw:Dict = {},
    ):
        
        traces = self.surveys3d_traces(
            wells=wells,
            area_unit=area_unit,
            depth_unit=depth_unit,
            mode=mode,
            trace_kw=trace_kw
        )
        
        layout = go.Layout(
            title = 'Surveys',
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **layout_kw
        )
        return go.Figure(data=traces,layout=layout)
        
    
    @validate_arguments
    def formations_vtk(
        self,
        wells: Optional[List[str]] = None,
        formations: Optional[List[str]] = None,
        to_crs:int = None,
        triangulate:bool = False
    ):
        coords = self.formations(wells = wells, formations = formations).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        data = {}
        for fm in coords['formation'].unique():
            _df = coords.loc[coords['formation']==fm,['x','y','tvdss_top']].values
            _surv = pv.PolyData(_df)
            if triangulate:
                _surv = _surv.delaunay_2d()
            data[fm] = _surv
        
        fm_blocks = pv.MultiBlock(data)
        return fm_blocks
    
    @validate_arguments
    def units_vtk(
        self,
        wells: Optional[List[str]] = None,
        units: Optional[List[str]] = None,
        to_crs:int = None,
        triangulate:bool = False
    ):
        coords = self.units(wells = wells, units = units).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        data = {}
        for fm in coords['unit'].unique():
            _df = coords.loc[coords['unit']==fm,['x','y','tvdss_top']].values
            _surv = pv.PolyData(_df)
            if triangulate:
                _surv = _surv.delaunay_2d()
            data[fm] = _surv
        
        fm_blocks = pv.MultiBlock(data)
        return fm_blocks
    
    @validate_arguments
    def formations3d_traces(
        self,
        wells: Optional[List[str]] = None,
        formations: Optional[List[str]] = None,
        to_crs:int = None,
        scatter: bool = False,
        triangulate:bool = False,
        **kwargs
    ):
        coords = self.formations(wells = wells, formations = formations).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        traces = []
        for fm in coords['formation'].unique():
            _df = coords.loc[coords['formation']==fm,['x','y','tvdss_top']]
            
            if triangulate:
                tri = Delaunay(_df[['x','y']].values)
                
                trace = go.Mesh3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    i = tri.simplices[:,0],
                    j = tri.simplices[:,1],
                    k = tri.simplices[:,2],
                    name = fm,
                    **kwargs
                )
                traces.append(trace)

            if scatter:
                scatter_trace = go.Scatter3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    mode = 'markers',
                    name = fm,
                )
                traces.append(scatter_trace)

        return traces
    
    @validate_arguments
    def units3d_traces(
        self,
        wells: Optional[List[str]] = None,
        units: Optional[List[str]] = None,
        to_crs:int = None,
        scatter:bool = False,
        triangulate:bool = False,
        **kwargs
    ):
        coords = self.units(wells = wells, units = units).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        traces = []
        for fm in coords['unit'].unique():
            _df = coords.loc[coords['unit']==fm,['x','y','tvdss_top']]
            
            if triangulate:
                tri = Delaunay(_df[['x','y']].values)
                
                
                trace = go.Mesh3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    i = tri.simplices[:,0],
                    j = tri.simplices[:,1],
                    k = tri.simplices[:,2],
                    name = fm,
                    **kwargs
                )
                traces.append(trace)
            
            if scatter:
                scatter_trace = go.Scatter3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    mode = 'markers',
                    name = fm
                )
                traces.append(scatter_trace)
                
        return traces
    
    @validate_arguments
    def perforations3d_traces(
        self,
        wells: Optional[List[str]] = None,
        perforations: Optional[List[str]] = None,
        to_crs:int = None,
        scatter:bool = False,
        triangulate:bool = False,
        **kwargs
    ):
        coords = self.perforations(wells = wells, perforations = perforations).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        traces = []
        for fm in coords['perforation'].unique():
            _df = coords.loc[coords['perforation']==fm,['x','y','tvdss_top']]
            
            if triangulate:
                tri = Delaunay(_df[['x','y']].values)
                
                
                trace = go.Mesh3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    i = tri.simplices[:,0],
                    j = tri.simplices[:,1],
                    k = tri.simplices[:,2],
                    name = fm,
                    **kwargs
                )
                traces.append(trace)
            
            if scatter:
                scatter_trace = go.Scatter3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    mode = 'markers',
                    name = fm
                )
                traces.append(scatter_trace)
                
        return traces
    
    @validate_arguments
    def tools3d_traces(
        self,
        wells: Optional[List[str]] = None,
        tools: Optional[List[str]] = None,
        to_crs:int = None,
        scatter:bool = False,
        triangulate:bool = False,
        **kwargs
    ):
        coords = self.tools(wells = wells, tools = tools).reset_index()
        
        if to_crs:
            coords = coords.to_crs(to_crs)
        
        coords['x'] = coords['geometry'].x
        coords['y'] = coords['geometry'].y
        
        traces = []
        for fm in coords['tool'].unique():
            _df = coords.loc[coords['tool']==fm,['x','y','tvdss_top']]
            
            if triangulate:
                tri = Delaunay(_df[['x','y']].values)
                
                
                trace = go.Mesh3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    i = tri.simplices[:,0],
                    j = tri.simplices[:,1],
                    k = tri.simplices[:,2],
                    name = fm,
                    **kwargs
                )
                traces.append(trace)
            
            if scatter:
                scatter_trace = go.Scatter3d(
                    x = _df['x'].values,
                    y = _df['y'].values,
                    z = _df['tvdss_top'].values,
                    mode = 'markers',
                    name = fm
                )
                traces.append(scatter_trace)
                
        return traces

    @validate_arguments
    def formations3d(
        self,
        wells: Optional[List[str]] = None,
        formations: Optional[List[str]] = None,
        to_crs:int = None,
        title:str = 'Formations',
        width:float = 700,
        height:float = 700,
        scatter:bool = False,
        **kwargs
    ):
        
        traces = self.formations3d_traces(
            wells = wells, 
            formations = formations, 
            to_crs = to_crs, 
            scatter=scatter,
            **kwargs
        )
        
        layout = go.Layout(
            title = title,
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **kwargs
        )
        
        return go.Figure(data=traces,layout=layout)
    
    @validate_arguments
    def units3d(
        self,
        wells: Optional[List[str]] = None,
        units: Optional[List[str]] = None,
        to_crs:int = None,
        title:str = 'units',
        width:float = 700,
        height:float = 700,
        scatter:bool = False,
        **kwargs
    ):
        
        traces = self.units3d_traces(
            wells = wells, 
            units = units, 
            to_crs = to_crs, 
            scatter = scatter,
            **kwargs
        )
        
        layout = go.Layout(
            title = title,
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **kwargs
        )
        
        return go.Figure(data=traces,layout=layout)

    @validate_arguments
    def perforations3d(
        self,
        wells: Optional[List[str]] = None,
        perforations: Optional[List[str]] = None,
        to_crs:int = None,
        title:str = 'perforations',
        width:float = 700,
        height:float = 700,
        scatter:bool = False,
        **kwargs
    ):
        
        traces = self.perforations3d_traces(
            wells = wells, 
            perforations = perforations, 
            to_crs = to_crs, 
            scatter = scatter,
            **kwargs
        )
        
        layout = go.Layout(
            title = title,
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **kwargs
        )
        
        return go.Figure(data=traces,layout=layout)   

    @validate_arguments
    def tools3d(
        self,
        wells: Optional[List[str]] = None,
        tools: Optional[List[str]] = None,
        to_crs:int = None,
        title:str = 'tools',
        width:float = 700,
        height:float = 700,
        scatter:bool = False,
        **kwargs
    ):
        
        traces = self.tools3d_traces(
            wells = wells, 
            tools = tools, 
            to_crs = to_crs, 
            scatter = scatter,
            **kwargs
        )
        
        layout = go.Layout(
            title = title,
            scene = {
                'xaxis': {'title': 'Easting'},
                'yaxis': {'title': 'Northing'},
                'zaxis': {'title': 'TVDss'}
            },
            width = width,
            height = height,
            **kwargs
        )
        
        return go.Figure(data=traces,layout=layout)     
            
    
    @validate_arguments
    def structural_view_vtk(
        self,
        wells: Optional[Union[bool,List[str]]] = False,
        formations: Optional[Union[bool,List[str]]] = False,
        units: Optional[Union[bool,List[str]]] = False,
        to_crs:int = None,
        triangulate_formations:bool = False,
        triangulate_units:bool = False
    ):
        
        blocks = pv.MultiBlock()
        
        if wells != False:
            surv = self.surveys_vtk(wells = wells, to_crs = to_crs)
            for s in surv.keys():
                blocks[s]=surv[s]
        
        if formations != False:
            form = self.formations_vtk(wells = wells, formations = formations, to_crs = to_crs, triangulate=triangulate_formations)
            for t in form.keys():
                blocks[t]=form[t]
        
        if units != False:
            unit = self.units_vtk(wells = wells, units = units, to_crs = to_crs, triangulate=triangulate_units)
  
            for u in unit.keys():
                blocks[u]=unit[u]

        return blocks
    
    @validate_arguments(config=dict(arbitrary_types_allowed=True))
    def from_omegadb(
        self,
        engine:Engine,
        wells:List[str]=None, 
        fields:List[str]=None,
        blocks:List[str]=None,
        basins:List[str]=None
    ):
        """
        Add wells information from the Postgres Database scuervo91/oilbase
        It uses the structure and the sintaxis implemented specifically in that database
        """
        
        well_heads_query = """
            select
                w.well as well,
                w.kbe as rte,
                w.surface_x as easting,
                w.surface_y as northing,
                w.epsg as epsg,
                f.field as field,
                bk.block as block,
                bs.basin as basin
            from
                list.wells w
            join
                list.fields f on w.field_id = f.id
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """
        
        well_surveys_query= """
            SELECT 
                w.well as well,
                s.md as md,
                s.inc as inc,
                s.azi as azi
            FROM
                list.surveys s
            JOIN
                list.wells w ON s.well_id = w.id
            JOIN
                list.fields f ON w.field_id = f.id
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """

        well_perforations_query= """
            SELECT 
                w.well as well,
                p.id as id,
                p.md_top as md_top,
                p.md_bottom as md_bottom
            FROM
                list.perforations p
            JOIN
                list.wells w ON p.well_id = w.id
            JOIN
                list.fields f ON w.field_id = f.id
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """
        
        well_formations_tops_query= """
            SELECT 
                w.well as well,
                fm.formation as formation,
                ft.md_top as md_top,
                ft.md_bottom as md_bottom
            FROM
                list.formations_tops ft
            JOIN
                list.formations fm ON ft.formation_id = fm.id
            JOIN
                list.wells w ON ft.well_id = w.id
            JOIN
                list.fields f ON w.field_id = f.id        
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """
        
        well_units_tops_query= """
            SELECT 
                w.well as well,
                un.unit as formation,
                ut.md_top as md_top,
                ut.md_bottom as md_bottom
            FROM
                list.units_tops ut
            JOIN
                list.units un ON ut.unit_id = un.id
            JOIN
                list.wells w ON ut.well_id = w.id
            JOIN
                list.fields f ON w.field_id = f.id        
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """
        
        well_perf_status_query= """
            select 
                w.well, 
                p.id, 
                p.md_top, 
                p.md_bottom, 
                pf.start_date, 
                pf.end_date, 
                pf.status 
            from 
                events.perforations_status pf 
            join 
                list.perforations p on pf.perforation_id = p.id 
            join 
                list.wells w on p.well_id = w.id
            join 
                list.fields f on w.field_id = f.id
            join
                list.blocks bk on f.block_id = bk.id
            join
                list.basins bs on bk.basin_id = bs.id
        """


        #Custom the query
        query_list = {
            'well_heads':well_heads_query,
            'well_surveys': well_surveys_query,
            'well_perforations':well_perforations_query,
            'well_formations_tops':well_formations_tops_query,
            'well_units_tops':well_units_tops_query,
            'well_perf_status':well_perf_status_query
        }

        wells_where_bool = False
        if wells is not None:
            assert isinstance(wells,(str,list))

            for i in query_list:
                query_list[i] = query_list[i] + f" {'and' if wells_where_bool else 'where'} w.well in {tuple(wells)}".replace(',)',')')
                
            wells_where_bool = True

        if fields is not None:
            assert isinstance(fields,(str,list))
            
            for i in query_list:
                query_list[i] = query_list[i] + f" {'and' if wells_where_bool else 'where'} f.field in {tuple(fields)}".replace(',)',')')
            wells_where_bool = True
            
        if blocks is not None:
            assert isinstance(blocks,(str,list))
            
            for i in query_list:
                query_list[i] = query_list[i] + f" {'and' if wells_where_bool else 'where'} bk.block in {tuple(blocks)}".replace(',)',')')
            wells_where_bool = True
            
        if basins is not None:
            assert isinstance(basins,(str,list))
            
            for i in query_list:
                query_list[i] = query_list[i] + f" {'and' if wells_where_bool else 'where'} f.basin in {tuple(basins)}".replace(',)',')')
            wells_where_bool = True



        # query from oilbase
        df_dict = {}
        for i in query_list:
            try:
                _df = pd.read_sql(query_list[i], con=engine)     
                df_dict[i] = _df
            except:
                print(f'no entered {i}')
                df_dict[i] = None
               
        wells_dict = {}

        for i, r in df_dict['well_heads'].iterrows():
            
            try:
                s = Survey(crs=r['epsg'])
                s.from_deviation(
                    df_dict['well_surveys'].loc[df_dict['well_surveys']['well']==r['well'],['md','azi','inc']], 
                    surface_easting=r['easting'],
                    surface_northing=r['northing'],
                    kbe=r['rte']
                )
            except Exception as e:
                print(f'survey {e}')
                s = None
            
            
            tops = Formations()
            try:
                tops_fm = Formations()
                tops_fm_df = df_dict['well_formations_tops'].loc[df_dict['well_formations_tops']['well']==r['well'],['formation','md_top','md_bottom']]
                tops_fm.from_df(tops_fm_df,name='formation')
                tops.add_formations(tops_fm.formations)
            except Exception as e:
                print(f'Excfm {e}')
            try:
                tops_un = Formations()
                tops_un_df = df_dict['well_units_tops'].loc[df_dict['well_units_tops']['well']==r['well'],['formation','md_top','md_bottom']]
                tops_un.from_df(tops_un_df,name='formation')
                tops.add_formations(tops_un.formations)
            except Exception as e:
                print(f'units {e}')


            perf = Perforations()
            try:
                perf.from_df(df_dict['well_perforations'].loc[df_dict['well_perforations']['well']==r['well'],:],name='id')
            except Exception as e:
                print(f'perf {e}')
            
            #try:
            perf.status_from_df(df_dict['well_perf_status'].loc[df_dict['well_perf_status']['well']==r['well'],:],name='id')
            #except Exception as e:
            #    print(f'perf status{e}')
            w = Well(
                name = r['well'],
                surface_coordinate = Point(r['easting'], r['northing']),
                rte=r['rte'],
                crs = int(r['epsg']),
                survey = s,
                formations = tops,
                perforations = perf,        
                field = r['field'],
                block = r['block'],
                basin = r['basin']                
            )
            
            wells_dict[w.name] = w

        return Wells(wells=wells_dict)


import numpy as np
import matplotlib.pyplot as plt
import pandas as pd

def coulomb(mu):
    """coulomb [summary]

    Parameters
    ----------
    mu : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    c = np.power(np.sqrt(np.power(mu,2) + 1) + mu, 2)
    return c

def shmin(sv,pp,mu):
    """shmin [summary]

    Parameters
    ----------
    sv : [type]
        [description]
    pp : [type]
        [description]
    mu : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    smin = ((sv - pp) / coulomb(mu)) + pp
    return smin

def shmax(sv,pp,mu):
    """shmax [summary]

    Parameters
    ----------
    sv : [type]
        [description]
    pp : [type]
        [description]
    mu : [type]
        [description]

    Returns
    -------
    [type]
        [description]
    """
    smax = coulomb(mu) * (sv - pp) + pp
    return smax 

def zobackogram(sv,pp,mu, ax=None):
    """zobackogram [summary]

    Parameters
    ----------
    sv : [type]
        [description]
    pp : [type]
        [description]
    mu : [type]
        [description]
    ax : [type], optional
        [description], by default None
    """
    ax = ax or plt.gca()  

    smin = shmin(sv,pp,mu)
    smax = shmax(sv,pp,mu)

    lines = []
    #45° line. slope=1
    L1 = pd.DataFrame({'shmin':[0,smax*1.1],'shmax':[0,smax*1.1]})
    lines.append(L1)

    #Shmin Line
    L2 = pd.DataFrame({'shmin':[smin,smin],'shmax':[smin,sv]})
    lines.append(L2)

    #Shmax Horizonal line
    L3 = pd.DataFrame({'shmin':[sv,smax],'shmax':[smax,smax]})
    lines.append(L3)

    #Shmax diagonal Line
    L4 = pd.DataFrame({'shmin':[smin,sv],'shmax':[sv,smax]})
    lines.append(L4)
    #Internal Lines
    L5 = pd.DataFrame({'shmin':[smin,sv,sv],'shmax':[sv,sv,smax]})
    lines.append(L5)

    co = ['black','darkblue','darkred','darkgreen','grey']
    ls = ['-','-','-','-','--',]
    wd = [3,2,2,2,1]
    lb = ['45° Line','Shmin Line','Shmax line','Shmax Strike slip','limits']

    for i, v in enumerate(lines):
        ax.plot(v['shmin'],v['shmax'],color = co[i], linestyle = ls[i], linewidth = wd[i])
    ax.set_xlabel('Shmin')
    ax.set_ylabel('Shmax')
    ax.legend(lb)
    ax.set_title('Stress State')


import matplotlib.pyplot as plt 
import pandas as pd
import numpy as np

def lithotrack(df:pd.DataFrame,
               codecols:list,
               percols:list, 
               dtick:bool=False, 
               lims:list=None, 
               codedict: dict=None, 
               fontsize=8, 
               ax=None,
               correlation: pd.DataFrame = None,
               grid_numbers : list = [11,51],
               steps: list  = None,
               corr_kw={}):
    """lithotrack [summary]

    Parameters
    ----------
    df : pd.DataFrame
        [description]
    codecols : list
        [description]
    percols : list
        [description]
    dtick : bool, optional
        [description], by default False
    lims : list, optional
        [description], by default None
    codedict : dict, optional
        [description], by default None
    fontsize : int, optional
        [description], by default 8
    ax : [type], optional
        [description], by default None
    correlation : pd.DataFrame, optional
        [description], by default None
    grid_numbers : list, optional
        [description], by default [11,51]
    steps : list, optional
        [description], by default None
    corr_kw : dict, optional
        [description], by default {}
    """
    lit=ax or plt.gca()
    
    def_corr_kw = {
    'color': 'red',
    'linestyle':'--',
    'linewidth': 2
    }    
    for (k,v) in def_corr_kw.items():
        if k not in corr_kw:
            corr_kw[k]=v
    
    df.index.names=['depth']
    df=df.reset_index()
    df=df.loc[(df.index>=lims[0])&(df.index<=lims[1]),:]
    #Create a pivot table concatenating the lithology code names
    mm=pd.DataFrame()
    for (k,v) in enumerate(codecols):    
        m=df.pivot_table(index=['depth'],columns=[v],values=percols[k])
        mm=pd.concat([mm,m],axis=1)

    #Merge in a single dataframe the repeated colnames
    mm=mm.fillna(0)
    lm=pd.DataFrame()
    for i in mm.columns.unique():
        if mm[i].ndim>1:             
            lm[i]=mm[i].max(axis=1)
        elif mm[i].ndim==1:
            lm[i]=mm[i]
    try:
        lm=lm.drop(columns=[0])
    except:
        pass
    lmc=np.cumsum(lm,axis=1)
    
    for i, col in enumerate(lmc.columns):
        lit.fill_betweenx(lmc.index, lmc.iloc[:,i], label=codedict[col], zorder=-i)
        
    if lims==None: #Depth Limits
        lims=[df.index.min(),df.index.max()]

    lit.set_ylim([lims[1],lims[0]])
        
    #Set the vertical grid spacing
    if steps is None:
        mayor_grid = np.linspace(lims[0],lims[1],grid_numbers[0])
        minor_grid = np.linspace(lims[0],lims[1],grid_numbers[1])
    else:
        mayor_grid = np.arange(lims[0],lims[1],steps[0])
        minor_grid = np.arange(lims[0],lims[1],steps[1])
        
    lit.legend()
    lit.set_xlim([0,100])
    lit.set_yticks(mayor_grid)
    lit.set_yticks(minor_grid,minor=True)  
    if dtick==True:
        lit.set_yticklabels(mayor_grid)
    else:
        lit.set_yticklabels([])
    lit.set_xlabel("Lithology")
    lit.xaxis.tick_top()
    lit.xaxis.set_label_position("top")
    lit.tick_params("both",labelsize=fontsize)    

def lithointtrack(depth,lint=None,dtick=False, lims=None, codedict=None, fontsize=8, ax=None):
    lth = ax or plt.gca()
    master=lint[(lint.index>=lims[0])&(lint.index<=lims[1])]
    m=pd.get_dummies(master)
    try: 
        m=m.drop(columns=[0])
    except:
        pass
    for i, col in enumerate(m.columns):
        lth.fill_betweenx(m.index, m.iloc[:,i], label=codedict[col], zorder=-i)
        
    if lims==None: #Depth Limits
        lims=[depth.max(),depth.min()]
        lth.set_ylim(lims)
    else:
        lth.set_ylim([lims[1],lims[0]])
    lth.legend()
    lth.set_xlim([0,1])
    lth.set_yticks(np.linspace(lims[0],lims[1],11))
    lth.set_yticks(np.linspace(lims[0],lims[1],51),minor=True)  
    if dtick==True:
        lth.set_yticklabels(np.linspace(lims[0],lims[1],11))
    else:
        lth.set_yticklabels([])
    lth.set_xlabel("Lithology Interpreted")
    lth.xaxis.tick_top()
    lth.xaxis.set_label_position("top")
    lth.tick_params("both",labelsize=fontsize)